//------------------------------------------------------------------------------------------------------
//
//                  PiRadio　NFM復調プログラム        Ver 1.10
//
//                  (C) 2016 Takashi Kato          Radiun.net
//
//
//                  Ver 1.00    2016/12/07      初版リリース
//                  Ver 1.01    2016/12/18      PLL設定d[2]の不具合修正
//                  Ver 1.10    2016/12/30      Emphasis追加、IIR-2次HPF追加、IIR-1次LPF追加、AFレベル計追加
//                                              FIR DCcut fc=750Hz を削除、IIR DCcut fc=100Hz に変更
//
//------------------------------------------------------------------------------------------------------

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include <alsa/asoundlib.h> // ALSA Audio

#include <X11/Xlib.h>       // XWindow
#include <X11/keysym.h>     // XWindow
#include <X11/Xutil.h>      // XWindow

#include <fcntl.h>          // GPIO
#include <sys/mman.h>       // GPIO
#include <unistd.h>         // GPIO

#define GPIO_PULLNONE 0x0   // GPIO
#define GPIO_PULLDOWN 0x1   // GPIO
#define GPIO_PULLUP   0x2   // GPIO
#define GPIO_INPUT    0x0   // GPIO
#define GPIO_OUTPUT   0x1   // GPIO

#define WIN_W     512       // XWindow
#define WIN_H     300       // XWindow
#define WIN_X     100       // XWindow
#define WIN_Y     100       // XWindow
#define BORDER    2         // XWindow

#define PERI_BASE     0x3F000000                // GPIO
#define GPIO_BASE     (PERI_BASE + 0x200000)    // GPIO
#define BLOCK_SIZE    4096                      // GPIO

static volatile unsigned int *Gpio;             // GPIO

char title[] = {"PiRadio V1.10"};
int x;

double re[1024];            // raw data I
double im[1024];            // raw data Q
double av[1024];
double i[1024];
double q[1024];
double coef[65];
double rfpow[512];          // RF Power
double afave;               // AF Average Power
double firi;
double firq;
double resi;
double resq;
double ofi=0;
double ofq=0;

float fpll=367.513;         // PLL Synth Freq /2
float fstep=0.001;          // Freq step MHz
int nwsw=0;                 // Narrow=0 Mid=1 Wide=2
int ofc=1;                  // Offset CAL ON=1/OFF=0
int resw=1;                 // Rotary Encoder Event
int lastmsb=0;              // Narrow Buffer MSB
int mode=0;                 // Mode FM:0 AM:1
int rxsw=0;                 // Speana/Demod SW
int emp=0;                  // Emphasis for NFM

//　IIR Coef
double a0[6], a1[6], a2[6], a3[6], a4[6];
double b0[6], b1[6], b2[6], b3[6], b4[6];

int rx_r, ry_r, wx_r, wy_r; // Mouse
unsigned int mask_r;        // Mouse

void GPIO_init();
void AD9218_wide();
void AD9218_narrow();
void ADF4351();
void FFT(double* real, double* imag, int length);
void gpio_init();
void gpio_configure_pull (int pin, int pullmode);
void gpio_configure(int pin, int mode);
void gpio_set(int pin);
void gpio_clear(int pin);
int  gpio_read (int pin);
void fir(double i, double q, int length);
void iirlpf2coef(int n, int fc, int smplrate);
void iirhpf2coef(int n, int fc, int smplrate);
void iirlpf1coef(int n, int fc, int smplrate);
float iir(int n, float vin);
void Fircoef(double* fcoef, int fc, int length);

int main( void )
{
    Display*       dpy;             // Display
    Window         root;            // Root Window
    Window         root_r, child_r; // Mouse
    Window         win;             // Display Window
    int            screen;          // Screen
    unsigned long  black,white;     // Pixcel value
    GC             gc;              // Graphic Context
    XEvent         evt;             // Event strunc

    Colormap       cmap;            // Color map
    XColor         color, exact;
    unsigned long  green, red;
    unsigned long  aqua, navy;
    
    KeySym keysym;

    FILE *fp;                       //  File pointer
    char sdata[128];

    double val;                     //　データ受信用
    
    int dr;
    int n;
    int m;
    int tx;
    int yy;
    int yi;
    int ly;
    int li;
    int lp;
    int yp;
    int vb;
    int xsize;
    int xstart;
    float xstep;
    int spa=0;
    int dem=0;
    int rcd=0;
    int afs=0;
    int xys=0;
    int bws=40;
    int bwf=40;
    double f;
    float vf=0;
    float bzero=1;
    
    int t;
    
    float fm;
    float lfm;
    float diff;
    float sout;
    float afpow;
    
    x=0;
    
    // ALSA サウンドデバイス設定
    const static char device[] = "hw:0";
    const static snd_pcm_format_t format = SND_PCM_FORMAT_S16_LE;
    const static snd_pcm_access_t access = SND_PCM_ACCESS_RW_INTERLEAVED;
    const static unsigned int sampling_rate = 48000;    // Sampling rate = 48kHz
    const static unsigned int channels = 1;             // Mono
    const static unsigned int soft_resample = 1;        // Re-sample authorized
    const static unsigned int latency = 100000;         // 100000us
    const unsigned int buffer_size = 512;
    
    snd_pcm_t *pcm;
    int16_t buffer[ buffer_size ];

    snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );
    snd_pcm_set_params(
                        pcm, format, access,
                        channels, sampling_rate, soft_resample, latency
                       );
    
    // GPIO 初期化
    GPIO_init();
    
    // XWindow 初期化
    dpy = XOpenDisplay( "" );
    root   = DefaultRootWindow( dpy );
    screen = DefaultScreen( dpy );
    cmap   = DefaultColormap( dpy, screen );

    // 白、黒のピクセル値
    white  = WhitePixel( dpy, screen );
    black  = BlackPixel( dpy, screen );

    // 緑、赤、アクア、ネービーのピクセル値
    XAllocNamedColor( dpy, cmap, "green", &color, &exact );
    green = color.pixel;

    XAllocNamedColor( dpy, cmap, "red", &color, &exact );
    red = color.pixel;
    
    XAllocNamedColor( dpy, cmap, "aqua", &color, &exact );
    aqua = color.pixel;
    
    XAllocNamedColor( dpy, cmap, "navy", &color, &exact );
    navy = color.pixel;
    
    // ウィンドウ作成
    win = XCreateSimpleWindow( dpy, root,
                            WIN_X, WIN_Y, WIN_W, WIN_H, BORDER, black, black);

    gc = XCreateGC( dpy, win, 0, NULL );
    
    //　イベントを受け取るマスクを指定
    XSelectInput( dpy, win, KeyPressMask | ExposureMask | ButtonPressMask);
    
    XMapWindow( dpy, win );
    
    // Main Loop
    while( 1 ) {
        while ( XPending( dpy ) ) {
            
            //　イベントがあった場合の処理
            XNextEvent( dpy, &evt );

            switch( evt.type ) {
                    
                //　立ち上げ時の処理
                case Expose:
                    if ( evt.xexpose.count == 0 ) {
                                
                        GPIO_init();                    // GPIO initial
                        ADF4351();                      // PLL initial
                        gpio_set(8);                    // nwsw=0;
                        nwsw=0;
                        iirlpf2coef(0, 20000, 48000);   // IIR LPF default fc=40kHz
                        iirlpf2coef(1, 20000, 48000);   // IIR LPF default fc=40kHz
                        iirlpf1coef(2, 300, 48000);     // Emphasis fc=300Hz -6dB/oct
                        iirhpf2coef(3, 100, 48000);     // DC cut fc=100Hz
                        iirhpf2coef(4, 100, 48000);     // DC cut fc=100Hz
                        iirhpf2coef(5, 100, 48000);     // IIR HPF fc=100Hz
                        Fircoef(coef, 27, 64);          // FIR LPF default 48kHz x 27/32 = 40kHz

                    }
                    break;
        
                case ButtonPress:
                    // クリックしたマウスの位置を中心周波数に設定
                    XQueryPointer(dpy, win,
                                  &root, &child_r,
                                  &rx_r, &ry_r, &wx_r, &wy_r,
                                  &mask_r);
                    fpll=fpll+1.5625*(wx_r-256)/512;
                    ADF4351();
                    break;
                    
                //　キー入力のイベント処理
                case KeyPress:
                    keysym = XKeycodeToKeysym(dpy, evt.xkey.keycode, 0);
                    switch(keysym){
                        case XK_Escape:
                            // リソースの解放
                            XFreeGC( dpy, gc );
                            XDestroyWindow( dpy, win );
                            XCloseDisplay( dpy );
                            snd_pcm_close( pcm );
                            return 0;
                            
                        case XK_0:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+0;
                            }else{
                                vf=vf+0*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_1:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+1;
                            }else{
                                vf=vf+1*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_2:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+2;
                            }else{
                                vf=vf+2*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_3:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+3;
                            }else{
                                vf=vf+3*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_4:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+4;
                            }else{
                                vf=vf+4*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_5:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+5;
                            }else{
                                vf=vf+5*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_6:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+6;
                            }else{
                                vf=vf+6*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_7:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+7;
                            }else{
                                vf=vf+7*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_8:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+8;
                            }else{
                                vf=vf+8*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_9:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+9;
                            }else{
                                vf=vf+9*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_Return:
                            // 周波数決定
                            if (vf>45){
                                fpll=vf;
                                ADF4351();
                                vf=0;
                                bzero=1;
                            }else{
                                vf=0;
                            }
                            break;
                            
                        case XK_period:
                            // Enter Below Zero
                            bzero=0.1;
                            break;
                            
                        case XK_F1:
                            //  BW=5kHz
                            bws=5;
                            iirlpf2coef(0, 2500, 48000);
                            iirlpf2coef(1, 2500, 48000);
                            break;
                            
                        case XK_F2:
                            //  BW=10kHz
                            bws=10;
                            iirlpf2coef(0, 5000, 48000);
                            iirlpf2coef(1, 5000, 48000);
                            break;
                            
                        case XK_F3:
                            //  BW=20kHz
                            bws=20;
                            iirlpf2coef(0, 10000, 48000);
                            iirlpf2coef(1, 10000, 48000);
                            break;
                            
                        case XK_F4:
                            //  BW=30kHz
                            bws=30;
                            iirlpf2coef(0, 15000, 48000);
                            iirlpf2coef(1, 15000, 48000);
                            break;
                            
                        case XK_F5:
                            //  BW=40kHz
                            bws=40;
                            iirlpf2coef(0, 20000, 48000);
                            iirlpf2coef(1, 20000, 48000);
                            break;
                            
                        case XK_F6:
                            //  BW=48kHz
                            bws=48;
                            break;
                            
                        case XK_F7:
                            //  BW=5kHz
                            bwf=5;
                            Fircoef(coef, 3, 64);          //  48kHz x 3/32 = 4.5kHz
                            break;
                            
                        case XK_F8:
                            //  BW=10kHz
                            bwf=10;
                            Fircoef(coef, 7, 64);          //  48kHz x 7/32 = 10.5kHz
                            break;
                            
                        case XK_F9:
                            //  BW=20kHz
                            bwf=20;
                            Fircoef(coef, 13, 64);         //  48kHz x 13/32 = 19.5kHz
                            break;
                            
                        case XK_F10:
                            //  BW=30kHz
                            bwf=30;
                            Fircoef(coef, 20, 64);         //  48kHz x 20/32 = 30.0kHz
                            break;
                            
                        case XK_F11:
                            //  BW=40kHz
                            bwf=40;
                            Fircoef(coef, 27, 64);         //  48kHz x 27/32 = 40.5kHz
                            break;
                            
                        case XK_F12:
                            //  BW=48kHz
                            bwf=48;
                            Fircoef(coef, 32, 64);         //  48kHz x 32/32 = 48.0kHz
                            break;
                            
                        case XK_Up:
                            //　周波数ステップ変更
                            if (fstep<100) {
                                fstep=fstep*10;
                            }
                            break;
                            
                        case XK_Down:
                            //　周波数ステップ変更
                            if (fstep>0.01) {
                                fstep=fstep/10;
                            }
                            break;
                            
                        case XK_Right:
                            // 周波数アップ
                            fpll=fpll+fstep;
                            ADF4351();
                            break;
                            
                        case XK_Left:
                            // 周波数ダウン
                            fpll=fpll-fstep;
                            ADF4351();
                            break;
                            
                        case XK_Control_L:
                            // Narrow-Mid-Wide
                            nwsw++;
                            if (nwsw>2) {
                                nwsw=0;
                            }
                            if (nwsw==0) {      // Narrow
                                gpio_clear(8);
                                gpio_clear(16);
                            }
                            if (nwsw==1) {      // Mid
                                gpio_set(8);
                                gpio_set(16);
                            }
                            if (nwsw==2) {      // Wide
                                gpio_set(8);
                                gpio_clear(16);
                            }
                            break;
                            
                        case XK_space:
                            // Speana/Demod Quick SW
                            if (rxsw==0) {
                                // Demod mode
                                rxsw=1;
                                dem=1;
                                spa=0;
                                nwsw=0;
                                gpio_clear(8);
                                gpio_clear(16);
                            }else{
                                // Speana mode
                                rxsw=0;
                                dem=0;
                                spa=1;
                                nwsw=1;
                                gpio_set(8);
                                gpio_set(16);
                            }
                            break;
                            
                        case XK_m:
                            // Mode FM / AM
                            if (mode==0) {
                                mode=1;
                            }else{
                                mode=0;
                            }
                            break;
                            
                        case XK_d:
                            // Demodulator
                            if (dem==0) {
                                dem=1;
                            }else{
                                dem=0;
                            }
                            break;
                            
                        case XK_e:
                            // Emphasis
                            if (emp==0) {
                                emp=1;
                            }else{
                                emp=0;
                            }
                            break;
                            
                        case XK_a:
                            // AF sound
                            if (afs==0) {
                                afs=1;
                            }else{
                                afs=0;
                            }
                            break;
                            
                        case XK_c:
                            // Offset CAL
                            if (ofc==0) {
                                ofc=1;
                            }else{
                                ofc=0;
                            }
                            break;
                            
                        case XK_x:
                            // XY scope
                            if (xys==0) {
                                xys=1;
                            }else{
                                xys=0;
                            }
                            break;
                            
                        case XK_Tab:
                            //　オシロ、スペアナ切り替え
                            if (spa==0) {
                                spa=1;
                            }else{
                                spa=0;
                            }
                            break;
                            
                        case XK_r:
                            //  Record start or stop
                            if (rcd==0) {
                                rcd=1;
                                 if ((fp=fopen("recorddata.csv", "w")) == NULL) {
                                    printf("file open error! \n");
                                }
                            }else{
                                rcd=0;
                                fclose(fp);
                            }
                            break;
                            
                    }
             }
        }
        
        //　ロータリエンコーダ処理
        if (gpio_read(13)==1 && resw==0)
        {
            fpll=fpll-fstep;
            resw=1;
            ADF4351();
        }
        if (gpio_read(19)==1 && resw==0)
        {
            fpll=fpll+fstep;
            resw=1;
            ADF4351();
        }
        if (gpio_read(13)==0 && gpio_read(19)==0) {
            resw=0;
        }
        
        //　広帯域・狭帯域　切替とWait
        if (nwsw>0) {
            //  Wide or Mid Band
            x=1;
            usleep(30000);
            AD9218_wide();
        }else{
            //  Narrow Band
            //　MSBが変化するまで待つ（I/Q切り替わりを待つ）
            if (gpio_read(5)==lastmsb) {
                //　復調動作を禁止、待機状態、画面表示を１回だけ実行
                x++;
            }else{
                //　復調動作を許可、MSB読み出しとADデータ取得
                x=0;
                lastmsb=gpio_read(5);
                AD9218_narrow();
            }
        }
        
        //  *******************************************************************************
        //　                                     復調処理
        //　                         画面表示と同じタイミングでは実行しない
        //  *******************************************************************************
        if (x==0) {
            
            afpow=0;
            
            for (m=0; m<512; m++) {
                
                //　受信データ　Ich  re[m]
                //　受信データ　Qch  im[m]
                
                fir(re[m], im[m], 64);              // FIR filter to compen Sinc
                
                resi=iir(3, resi);                  // DC cut Ich
                resq=iir(4, resq);                  // DC cut Qch
                
                if (bws<48) {
                    resi=iir(0, resi);              // IIR filter Ich
                    resq=iir(1, resq);              // IIR filter Qch
                }
                
                // ------------------------------------------------------------------------
                // ----------------------------  Narrow FM 復調 ----------------------------
                // ------------------------------------------------------------------------
                if (mode==0) {
                    fm=atan2f(resi,resq);           // ArcTan -2pi to 2pi
                    diff = fm-lfm;                  // Differential
                    if (fabs(diff)<M_PI) {
                        sout = diff;
                    }else{
                        if(diff>0) {
                            sout = diff - 2 * M_PI;
                        }else{
                            sout = diff + 2 * M_PI;
                        }
                    }
                    if (emp==1) {                   // Emphasis mode ON
                        sout=0.1*iir(2,sout);       // Emphasis fc=300Hz 6dB/oct
                        sout=iir(5,sout);           // HPF fc=100Hz
                        afpow=afpow+fabs(sout);
                    }
                }
                
                // ------------------------------------------------------------------------
                // --------------------------------  AM 復調 -------------------------------
                // ------------------------------------------------------------------------
                if (mode==1) {
                    sout=sqrtf(resi*resi+resq*resq)*100;
                }
                
                buffer[m] = (int)(sout * 10000);    // ALSA sound buffer +/-3.14 x 10000
                lfm=fm;
                
                if (afs==0) {
                    //　FIR処理後の波形を画像に出力
                    re[m]=resi;
                    im[m]=resq;
                }else{
                    //　AF信号波形を画像に出力
                    re[m]=sout*0.1;
                    im[m]=sout*0.1;
                }
                
            }
            
            // ----------------------------------------------------------------------------
            // -------------------------  ALSAバッファへ音声出力 -----------------------------
            // ----------------------------------------------------------------------------
            if (dem==1 && spa==0) {
                int write_result = snd_pcm_writei ( pcm, ( const void* )buffer, buffer_size );
                if( write_result < 0 ) {
                    if( snd_pcm_recover( pcm, write_result, 0 ) < 0 ) {
                        printf( "Unable to recover this stream." );
                        snd_pcm_close( pcm );
                        abort();
                    }
                }
            }
        }
        
        //  *******************************************************************************
        //　                                     画面表示
        //　                         復調処理と同じタイミングでは実行しない
        //  *******************************************************************************
        if (x==1) {
            
            //  Record data
            if (rcd==1) {
                sprintf(sdata, "%f \n", rfpow[1]);
                fputs(sdata, fp);
            }
            
            //　FFT処理
            if (spa==1) FFT(re, im, 1024);
        
            //　画面のクリア
            XClearWindow(dpy, win);
            
            // 描画色にnavyを設定
            XSetForeground( dpy, gc, navy );
            
            for (n=0; n<11; n++){
                XDrawLine( dpy, win, gc, 6, n*30, 506, n*30);
                XDrawLine( dpy, win, gc, 6+n*50, 0, 6+n*50, 300);
            }
            // 描画色にaquaを設定
            XSetForeground( dpy, gc, aqua );
            
            if (spa==1) {
                //　スペアナ画面設定
                if (nwsw==0) {      // Narrow
                    xstart=0;
                    xsize=1024;
                    xstep=0.5;
                }else{              // Wide
                    xstart=96;
                    xsize=1024;
                    xstep=0.5;
                }
            }else{
                //　オシロ画面設定
                xstart=1;
                xsize=512;
                xstep=1;
            }
            
            // RF power バッファデータのシフト
            if (spa==0) {
                for (n=0; n<xsize; n++) {
                    rfpow[n]=rfpow[n+1];
                }
                rfpow[511]=0;
            }
            
            for (n=1; n<xsize; n++) {
                
                // Calc RF power
                if (spa==0) {
                    rfpow[511]=rfpow[511]+sqrtf(pow(re[n],2)+pow(im[n],2));
                }
                
                // 値を座標に変換
                if (spa==0) {
                    //　オシロ表示
                    if (xys==0) {
                        yy=75-re[n]*60;
                        yi=225-im[n]*60;
                        yp=150-rfpow[n]*1;
                    }else{
                        yy=150-re[n]*100;
                        yi=256-im[n]*100;
                    }
                }else{
                    //　FFT表示
                    yy=150-2*20*log10(sqrtf((re[n]*re[n]+im[n]*im[n])));
                    av[n]=0.9*av[n]+0.1*yy;     // Averaging x10
                    yy=av[n];
                }
                
                //　画面端処理
                if (yy<  0) yy=  0;
                if (yy>300) yy=300;
                if (yi<  0) yi=  0;
                if (xys==0) {
                    if (yi>300) yi=300;
                }else{
                    if (yi>511) yi=511;
                }
                
                //　オシロ波形描画
                if (n>1 && spa==0) {
                    if (xys==0) {
                        XDrawLine( dpy, win, gc, (n-1)*xstep, ly, n*xstep, yy);
                        XDrawLine( dpy, win, gc, (n-1)*xstep, li, n*xstep, yi);
                        XDrawLine( dpy, win, gc, (n-1)*xstep, lp, n*xstep, yp);

                    }else{
                        //　XY表示
                        XDrawLine( dpy, win, gc, li, ly, yi, yy);
                    }
                }
                
                //　スペアナ波形表示
                if (n>1 && n<xsize/2 && spa==1) {
                    XDrawLine( dpy, win, gc, (xsize/2+n-1)*xstep, ly, (xsize/2+n)*xstep, yy);
                }
                if (n+1>xsize/2 && spa==1) {
                    XDrawLine( dpy, win, gc, (-xsize/2+n-1)*xstep, ly, (-xsize/2+n)*xstep, yy);
                }
                
                ly=yy;
                li=yi;
                lp=yp;
                
            }
            
            //　テキスト表示
            char txt[55];
            if (vf>0) fpll=vf;
            if (nwsw==0 && spa==1) sprintf(txt, "Center=%8.3f MHz  Span= 48 kHz  Step=%8.3f MHz", fpll, fstep);
            if (nwsw==1 && spa==1) sprintf(txt, "Center=%8.3f MHz  Span= 1.5 MHz Step=%8.3f MHz", fpll, fstep);
            if (nwsw==2 && spa==1) sprintf(txt, "Center=%8.3f MHz  Span= 50 MHz  Step=%8.3f MHz", fpll, fstep);
            if (nwsw==0 && spa==0) sprintf(txt, "Center=%8.3f MHz  Time 10ms/div Step=%8.3f MHz", fpll, fstep);
            if (nwsw==1 && spa==0) sprintf(txt, "Center=%8.3f MHz  Time 32us/div Step=%8.3f MHz", fpll, fstep);
            if (nwsw==2 && spa==0) sprintf(txt, "Center=%8.3f MHz  Time 1us/div  Step=%8.3f MHz", fpll, fstep);
            XDrawString(dpy, win, gc, 20, 20, txt, 52);
            if (mode==0) sprintf(txt, "Mode: FM  IIR= %2d kHz FIR= %2d kHz",bws, bwf);
            if (mode==1) sprintf(txt, "Mode: AM  IIR= %2d kHz FIR= %2d kHz",bws, bwf);
            XDrawString(dpy, win, gc, 20, 40, txt, 33);
            if (emp==1)  {
                afave=0.9*afave+0.1*afpow/512;
                sprintf(txt, "NFM Empasis ON AF=%5.1f dB",20*log10f(afave));
                XDrawString(dpy, win, gc, 20, 60, txt, 26);
            }
        }
    }
}

//-------------------------------------------------------------------
//                          FIR Filter
//-------------------------------------------------------------------
void fir(double i, double q, int length)
{
    int k;
    double io, qo;
    
    static double zi[1024];
    static double zq[1024];
    
    int size;
    
    size=length;
    
    io=0;
    qo=0;
    
    for (k=0; k<size-1; k++) {
        zi[size-1-k] = zi[size-2-k];
        io = io + coef[size-1-k] * zi[size-1-k];
        
        zq[size-1-k] = zq[size-2-k];
        qo = qo + coef[size-1-k] * zq[size-1-k];
    }
    zi[0] = i;
    io = io + coef[0] * zi[0];
    
    zq[0] = q;
    qo = qo + coef[0] * zq[0];
    
    resi=io;
    resq=qo;

}

//-------------------------------------------------------------------
//                          A/Dコンバータ受信　50Msps 1.56Msps
//-------------------------------------------------------------------
void AD9218_wide()
{
    double val;
    double avi=0;
    double avq=0;
    int v;
    int lv;
    int gv=0;
    int n;
    int m;
    
    gpio_set(11);       // sce on
    gpio_set(6);        // Ich select
    
    //  Ich Load
    for (m=0; m<1024; m++) {
        
        gv=0;
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read(9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;          //  Comp on 2
        val=(double)gv/(double)511;
        re[m]=val;
        
        avi=avi+val;
    }
    
    gpio_clear(6);       // Qch select
    
    //  Qch Load
    for (m=0; m<1024; m++) {
        
        gv=0;
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read(9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;          //  Comp on 2
        val=(double)gv/(double)511;
        im[m]=val;
        
        avq=avq+val;
    }
    
    if (ofc==1) {                       //  CAL Offset ON/OFF
        ofi=avi/1024;
        ofq=avq/1024;
    }
    
    for (m=0; m<1024; m++) {            // Remove Offset
        re[m]=re[m]-ofi;
        im[m]=im[m]-ofq;
    }
    
    gpio_clear(11);                     //  CE0 0
    
}

//-------------------------------------------------------------------
//                          A/Dコンバータ受信　48ksps
//-------------------------------------------------------------------
void AD9218_narrow()
{
    double val;
    int v;
    int lv;
    int gv=0;
    int n;
    int m;
    
    gpio_set(11);       // sce on
    gpio_set(6);        // Ich select
    
    //  Ich Load
    for (m=0; m<512; m++) {
        
        gv=0;
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read(9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        gpio_clear(7);         //  CK 0
        gpio_clear(7);         //  CK 0
        gpio_clear(7);         //  CK 0
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;          //  Comp on 2
        val=(double)gv/(double)511;
        re[m]=val;
        
    }
    
    
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off
    
    gpio_clear(6);      // Qch select
    gpio_set(11);       // sce on

    
    //  Qch Load
    for (m=0; m<512; m++) {
        
        gv=0;
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read(9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;          //  Comp on 2
        val=(double)gv/(double)511;
        im[m]=val;
        
    }
    
    gpio_clear(11);                     //  CE0 0
    
}

//-------------------------------------------------------------------
//                        シンセサイザ　ADF4531　設定
//-------------------------------------------------------------------
void ADF4351()
{
    int n;
    int m;
    int d[6];
    int b;
    
    float fo;
    float fstp;
    int fint;
    int div;
    int frac;
    int mod;
    int mmod;
    
    int i1,i2,i3,i4,m1,m2;
    
    float floop=25.000;     // Freq loopfilter=25.0MHz
    
    d[0]=0x00524E28;
    d[1]=0x0800E1A9;
    d[2]=0x00004E42;
    d[3]=0x000004B3;
    d[4]=0x009C803C;
    d[5]=0x00580005;
    
    fo=fpll*2;
    div=pow(2,int(log2(4400/fo)));
    fstp=0.008/div;
    fint=int(fo*div/floop);
    mod=int(floop/(fstp*div));
    frac=int(mod*(fo*div/floop-fint));
    
//    printf("\n");
//    printf("fset=%5.2f MHz\n",fpll);
//    printf("fosc=%5.2f MHz\n",fo);
//    printf("Int=%5d \n",fint);
//    printf("Div=%2d \n",div);
//    printf("Frac=%4d \n",frac);
//    printf("Mod=%4d \n",mod);
//    printf("fvco=%5.2f MHz\n",fo*div);
    
    d[0]=fint*0x8000+frac*8;
    d[1]=0x08008001+mod*8;
    d[4]=0;
    d[4]=d[4]+0x800000 * 1;             //  Feedback 0:divided 1:fundamental
    d[4]=d[4]+0x100000 * log2(div);     //  RF Divider /1 to /64
    d[4]=d[4]+0x001000 * 0xC8;          //  Clock Divider
    d[4]=d[4]+0x000800 * 0;             //  VCO Power Down 0:UP 1:DOWN
    d[4]=d[4]+0x000400 * 0;             //  Mute Til Lock 0:Disable 1:Enable
    d[4]=d[4]+0x000200 * 0;             //  Aux output select 0:divided 1:fundamental
    d[4]=d[4]+0x000100 * 0;             //  Aux Out 0:Disable 1:Enable
    d[4]=d[4]+0x000040 * 0;             //  Aux Out 0:-4dBm 1:-1dBm 2:+2dBm 3:+5dBm
    d[4]=d[4]+0x000020 * 1;             //  RF Out 0:Disable 1:Enable
    d[4]=d[4]+0x000008 * 3;             //  RF Out 0:-4dBm 1:-1dBm 2:+2dBm 3:+5dBm
    d[4]=d[4]+0x000001 * 4;             //  Control bit 4

    gpio_clear(12);                // LE L
    gpio_clear(2);                 // CK L
    gpio_clear(3);                 // DATA L
    usleep(100);
    
    for (m=0; m<6; m++) {
        
        gpio_clear(12);            // LE L
        usleep(100);
        
        for (n=0; n<32; n++) {
            b=d[m] & (int)pow(2,31-n);
            if (n==0) b=0;
            
            usleep(100);
            gpio_clear(2);         // CK L
            
            if (b==0) {
                gpio_clear(3);     // DATA H
            }else{
                gpio_set(3);       // DATA L
            }
            
            usleep(100);
            gpio_set(2);           // CK H
        }
        
        usleep(100);
        gpio_clear(2);             // CK L
        usleep(100);
        gpio_clear(3);             // DATA L
        usleep(100);
        gpio_set(12);              // LE H
        usleep(100);
        
    }
    
    gpio_clear(12);                // LE L
    gpio_clear(2);                 // CK L
    gpio_clear(3);                 // DATA L
    
//    printf("\n");
//    printf("%8x \n",d[0]);
//    printf("%8x \n",d[1]);
//    printf("%8x \n",d[2]);
//    printf("%8x \n",d[3]);
//    printf("%8x \n",d[4]);
//    printf("%8x \n",d[5]);

}

//-------------------------------------------------------------------
//                          GPIO　初期化
//-------------------------------------------------------------------
void GPIO_init()
{
    
    gpio_init();                            //
        
    gpio_configure(2,  GPIO_OUTPUT);        //  PLL CK
    gpio_configure(3,  GPIO_OUTPUT);        //  PLL DA
    gpio_configure(4,  GPIO_INPUT);         //  dummy
    gpio_configure(5,  GPIO_INPUT);         //  msbsw
    gpio_configure(6,  GPIO_OUTPUT);        //  iqsw
    gpio_configure(7,  GPIO_OUTPUT);        //  SCLK
    gpio_configure(8,  GPIO_OUTPUT);        //  nwsw
    gpio_configure(9,  GPIO_INPUT);         //  dat1
    gpio_configure(10, GPIO_INPUT);         //  dat2
    gpio_configure(11, GPIO_OUTPUT);        //  SCE
    gpio_configure(12, GPIO_OUTPUT);        //  PLL LE
    gpio_configure(13, GPIO_INPUT);         //  Rotary Encoder A
    gpio_configure(14, GPIO_INPUT);         //  dat9
    gpio_configure(15, GPIO_INPUT);         //  dat8
    gpio_configure(16, GPIO_OUTPUT);        //  Mid/Wide Band Select
    gpio_configure(17, GPIO_INPUT);         //  dat7
    gpio_configure(18, GPIO_INPUT);         //  pwm
    gpio_configure(19, GPIO_INPUT);         //  Rotary Encoder B
    gpio_configure(20, GPIO_INPUT);         //  GPIO rasp8
    gpio_configure(21, GPIO_INPUT);         //  GPIO rasp9
    gpio_configure(22, GPIO_INPUT);         //  dat5
    gpio_configure(23, GPIO_INPUT);         //  dat4
    gpio_configure(24, GPIO_INPUT);         //  dat3
    gpio_configure(25, GPIO_INPUT);         //  dat0
    gpio_configure(26, GPIO_INPUT);         //  GPIO rasp7
    gpio_configure(27, GPIO_INPUT);         //  dat6
    
    //gpio_configure_pull(2, GPIO_PULLDOWN);
    //gpio_configure_pull(3, GPIO_PULLDOWN);
    //gpio_configure_pull(4, GPIO_PULLDOWN);
    gpio_configure_pull(5, GPIO_PULLDOWN);
    //gpio_configure_pull(6, GPIO_PULLDOWN);
    //gpio_configure_pull(7, GPIO_PULLDOWN);
    //gpio_configure_pull(8, GPIO_PULLDOWN);
    gpio_configure_pull(9, GPIO_PULLDOWN);
    gpio_configure_pull(10, GPIO_PULLDOWN);
    //gpio_configure_pull(11, GPIO_PULLDOWN);
    gpio_configure_pull(12, GPIO_PULLDOWN);
    gpio_configure_pull(13, GPIO_PULLDOWN);
    gpio_configure_pull(14, GPIO_PULLDOWN);
    gpio_configure_pull(15, GPIO_PULLDOWN);
    //gpio_configure_pull(16, GPIO_PULLDOWN);
    gpio_configure_pull(17, GPIO_PULLDOWN);
    gpio_configure_pull(18, GPIO_PULLDOWN);
    gpio_configure_pull(19, GPIO_PULLDOWN);
    gpio_configure_pull(20, GPIO_PULLDOWN);
    gpio_configure_pull(21, GPIO_PULLDOWN);
    gpio_configure_pull(22, GPIO_PULLDOWN);
    gpio_configure_pull(23, GPIO_PULLDOWN);
    gpio_configure_pull(24, GPIO_PULLDOWN);
    gpio_configure_pull(25, GPIO_PULLDOWN);
    gpio_configure_pull(26, GPIO_PULLDOWN);
    gpio_configure_pull(27, GPIO_PULLDOWN);
    

}

void FFT(double* real, double* imag, int length)
{
    //　IQによる帯域2倍に対応
    
    int i, j, k, l;
    float m, h, g, q, p, a, b;
    
    float swap;
    
    //********************* 波形読込と同時にハミング窓を掛ける *******************
    
    for (i=0; i<=length; i++) {
        a = real[i] * (0.54 - 0.46 * cos(M_PI * (i-length) / length));          //Hamming
        b = imag[i] * (0.54 - 0.46 * cos(M_PI * (i-length) / length));          //Hamming
        
        real[i] = b;
        imag[i] = a;
        real[length-i] = a;
        imag[length-i] = b;
    }
    
    //******************************* FFTメイン **********************************
    m = log2(length);
    
    l = length;
    h = 1;
    
    for (g=1; g<=m; g++) {
        l = l / 2;
        k = 0;
        for (q=1; q<=h; q++) {
            p = 0;
            for (i=k; i<=l+k-1; i++) {
                j = i + l;
                a = real[i] - real[j];
                b = imag[i] - imag[j];
                real[i] = real[i] + real[j];
                imag[i] = imag[i] + imag[j];
                if (p==0){
                    real[j] = a;
                    imag[j] = b;
                }else{
                    real[j] = a * cos(2 * M_PI * p / length) + b * sin(2 * M_PI * p / length);
                    imag[j] = b * cos(2 * M_PI * p / length) - a * sin(2 * M_PI * p / length);
                }
                p = p + h;
            }
            k = k + l + l;
        }
        h = h + h;
    }
    
    j = length / 2;
    
    for (i=1; i<=length-1; i++) {
        k = length;
        if (j<i){
            swap = real[i];
            real[i] = real[j];
            real[j] = swap;
            swap = imag[i];
            imag[i] = imag[j];
            imag[j] = swap;
        }
        
        do{
            k = k / 2;
            j = j - k;
            if (k==0 && j==0) break;
        }while ((j+k)>=k);
        
        j = j + k + k;
    }

}


void iirlpf2coef(int n, int fc, int smplrate)
{
    //-------------------- IIR フィルタ係数生成 ------------------------
    int ws = smplrate;      //サンプリング周波数
    int w1;                 //カットオフ周波数
    float w;
    
    w1=fc;
    w = tanf(M_PI*w1/ws);
    
    a1[n] = -(2 * w*w - 2) / (1 + sqrtf(2) * w + w*w);
    a2[n] = -(1 - sqrtf(2) * w + w*w) / (1 + sqrtf(2) * w + w*w);
    b0[n] = w*w / (1 + sqrtf(2) * w + w*w);
    b1[n] = (2 * w*w) / (1 + sqrtf(2) * w + w*w);
    b2[n] = w*w / (1 + sqrtf(2) * w + w*w);
    //---------------------------------------------------------------
    
}

void iirhpf2coef(int n, int fc, int smplrate)
{
    //-------------------- IIR フィルタ係数生成 ------------------------
    int ws = smplrate;      //サンプリング周波数
    int w1;                 //カットオフ周波数
    float w;
    
    w1=fc;
    w = tanf(M_PI*w1/ws);
    
    a1[n] = -(2 * w - 2) / (1 + sqrtf(2)*w + w*w);
    a2[n] = -(1 - sqrtf(2)*w + w*w) / (1 + sqrtf(2)*w + w*w);
    b0[n] = 1 / (1 + sqrtf(2)*w + w*w);
    b1[n] = -2 / (1 + sqrtf(2)*w + w*w);
    b2[n] = 1 / (1 + sqrtf(2)*w + w*w);
    //---------------------------------------------------------------
    
}

void iirlpf1coef(int n, int fc, int smplrate)
{
    //-------------------- IIR フィルタ係数生成 ------------------------
    int ws = smplrate;      //サンプリング周波数
    int w1;                 //カットオフ周波数
    float w;
    
    w1=fc;
    w = tanf(M_PI*w1/ws);
    
    a1[n] = (1 - w) / (1 + w);
    a2[n] = 0;
    b0[n] = 1 / (1 + w);
    b1[n] = 1 / (1 + w);
    b2[n] = 0;
    //---------------------------------------------------------------
    
}

//-------------------------------------------------------------------
//                          IIR Filter
//-------------------------------------------------------------------
float iir(int n, float vin)
{
    static float z21[6];
    static float z22[6];
    float vo;
    
    vin = vin + a1[n] * z21[n];
    vin = vin + a2[n] * z22[n];
    vo = b0[n] * vin;
    vo = vo + b1[n] * z21[n];
    vo = vo + b2[n] * z22[n];
    z22[n] = z21[n];
    z21[n] = vin;
    
    return vo;
}

//-------------------------------------------------------------------
//                       Sinc補正 FIR係数の生成
//-------------------------------------------------------------------
void Fircoef(double* fcoef, int fc, int length)
{
    double fr[1024];
    double fj[1024];
    int n;
    int half;
    
    half = length/2;
    
    //  周波数特性を定義
    for (n=0; n<=half; n++) {
        if (n<fc) {
            fr[n]=1;
            fr[length-n]=1;
        }else{
            fr[n]=0;
            fr[length-n]=0;
        }
        fj[n]=0;
    }
    
    //  Sinc特性の補正
    const double N = 4;
    const double M = 32;
    double K;
    for (n=0; n<half; n++) {
        K = N*sinf(M_PI*(double)n*0.75/1563)/sinf(M*M_PI*(double)n*0.75/1563) / N*M;
        //fr[n]           =K*K*fr[n];
        //fr[length-n]    =K*K*fr[length-n];
    }
    
    //  DCカット
//    fr[0]=0;
//    fr[length]=0;
    
//    for (n=0; n<=length; n++) {
//        printf( "%d %6.5f \n",n,fr[n] );
//    }

    //  iFFTして係数を得る
    FFT(fr, fj, length);
    
    //  係数をシフトする
    for (n=0; n<half; n++) {
        fcoef[half+n]   =fr[n];
        fcoef[n]        =fr[half+n];
    }
    
//    for (n=0; n<=length; n++) {
//        printf( "FFT coef %d %6.5f \n",n,fcoef[n] );
//    }
    
    //　窓関数
    //for (n=0; n<half+1; n++) {
    //    fcoef[half+n] = fcoef[half+n] * (0.5*cosf(M_PI*(float)(n/half))+0.5);
    //    fcoef[half-n] = fcoef[half-n] * (0.5*cosf(M_PI*(float)(n/half))+0.5);
    //}
}

unsigned int * get_base_address()
{
    int fd = open("/dev/mem", O_RDWR | O_SYNC);
    if(fd < 0){
        printf("cannot open /dev/mem\n"); exit(-1);
    }
    
    #define PAGE_SIZE (4096)
    void * mmaped = mmap(NULL,
                         PAGE_SIZE,
                         PROT_READ | PROT_WRITE,
                         MAP_SHARED,
                         fd,
                         0x3F200000);
    if (mmaped < 0) { printf("mmap failure\n");
        exit(-1); }
    
    close(fd);
    return (unsigned int *) mmaped;
    
}

void gpio_init()
{
    if (Gpio) return;
    int fd;
    void *gpio_map;
    fd = open("/dev/mem", O_RDWR | O_SYNC);
    gpio_map = mmap(NULL, BLOCK_SIZE,
                    PROT_READ | PROT_WRITE, MAP_SHARED,
                    fd, GPIO_BASE );
    close(fd);
    Gpio = (unsigned int *) gpio_map;
}

void gpio_configure(int pin, int mode)
{
    int index = pin / 10;
    unsigned int mask = ~(0x7 << ((pin % 10) * 3));
    Gpio[index] = (Gpio[index] & mask) | ((mode & 0x7) << ((pin % 10) * 3));
}

void gpio_configure_pull (int pin, int pullmode)
{
    Gpio[37] = pullmode & 0x3;
    usleep(1);
    Gpio[38] = 0x1 << pin;
    usleep(1);
    Gpio[37] = 0;
    Gpio[38] = 0;
}

void gpio_set(int pin)
{
    Gpio[7] = 0x1 << pin;
}

void gpio_clear(int pin)
{
    Gpio[10] = 0x1 << pin;
}

int gpio_read (int pin)
{
    return (Gpio[13] & (0x1 << pin)) != 0;
}

