//-------------------------------------------------------------------------------------------------------------
//
//                  PiRadio　レシーバープログラム      Ver 2.30
//
//                  (C) 2016-2017 Takashi Kato  Radiun.net
//
//
//                  Ver 1.00    2016/12/07      初版リリース
//                  Ver 1.01    2016/12/18      PLL設定d[2]の不具合修正
//                  Ver 1.10    2016/12/30      Emphasis追加、IIR-2次HPF追加、IIR-1次LPF追加、AFレベル計追加
//                                              FIR DCcut fc=750Hz を削除、IIR DCcut fc=100Hz に変更
//                  Ver 1.11    2017/01/07      FFT不具合修正、窓関数 Hann/Hamming 選択機能追加、FIR DCcut に戻す
//					Ver 1.20    2017/01/29      GUIを大幅変更、Control Windowを追加、カーソル選択メニューを追加
//                                              FIR DCcutを削除、低CutoffのHPFに変更、オシロ、スペアナの設定変更を追加
//                                              マウスホイールを周波数変更に対応、画面サイズの変更に対応
//                  Ver 1.21    2017/02/12      CQエレクトロニクスセミナーで実演したスペクトラム拡散に対応
//                  Ver 1.22    2017/03/05      PLLシンセサイザ ADF4351 の高速ロックと低ノイズ化対応
//                  Ver 2.00    2017/04/14      GUI大幅変更（全ての操作をマウスで可能にする）
//                                              FIRフィルタ係数生成関数のデバッグ、IIR&FIRフィルタ周波数設定の細分化
//                                              波形、スペクトラムレベル表示をANT入力レベルに合わせる補正テーブルを追加
//                                              スケルチ機能＆スキャン機能追加、復調モードにDSB追加、PLLのusleep削除
//                                              コンパイル時 XkeycodeのWarning対策、ALSA underrun occurred 対策
//                                              FFT窓関数のデバッグ、トップスイッチ追加（画面停止、メモリ機能、波形保存）
//                                              絶対レベル（dBm）表示、エンファシスf特調整、波形ファイル保存追加
//                                              BW=48.8kHzとBW=1.56MHzのスペアナ表示にCICフィルタf特補正を追加
//                  Ver 2.10    2017/05/04      SSB受信用の改良（シンセのループフィルタ定数変更を推奨）
//                                              周波数ステップ最小100Hz化、DSBモードにゲイン調整追加
//                  Ver 2.20    2017/05/13      FPGA CICのATT機能追加に伴ってATT 0dB/12dB 切り替えを追加
//                                              スペアナ画面クリックで周波数設定モード（func=0）へ強制移行
//                  Ver 2.30    2017/05/18      スペアナのスパンを9段階に細分化、オシロの時間軸を9段階に細分化
//                                              FFT窓関数にBlackman-Harrisを追加、窓関数のデバッグ
//                                              Narrowで発生するパルス性ノイズをスペアナのみ読出タイミングをずらして対策
//
//-------------------------------------------------------------------------------------------------------------

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>       // Timer

#include <alsa/asoundlib.h> // ALSA Audio

#include <X11/Xlib.h>       // XWindow
#include <X11/keysym.h>     // XWindow
#include <X11/Xutil.h>      // XWindow
#include <X11/XKBlib.h>     // for XkbKeycodeToKeysym

#include <fcntl.h>          // GPIO
#include <sys/mman.h>       // GPIO
#include <unistd.h>         // GPIO

#define GPIO_PULLNONE 0x0   // GPIO
#define GPIO_PULLDOWN 0x1   // GPIO
#define GPIO_PULLUP   0x2   // GPIO
#define GPIO_INPUT    0x0   // GPIO
#define GPIO_OUTPUT   0x1   // GPIO

#define WIN_X     100       // XWindow
#define WIN_Y     100       // XWindow
#define BORDER    2         // XWindow

#define PERI_BASE     0x3F000000                // GPIO
#define GPIO_BASE     (PERI_BASE + 0x200000)    // GPIO
#define BLOCK_SIZE    4096                      // GPIO

static volatile unsigned int *Gpio;             // GPIO

char title[] = {"PiRadio Receiver V2.30"};
int x;
int WIN_W=500;              // XWindow
int WIN_H=500;              // XWindow

float fpll=123.000;         // PLL Synth Freq /2
float fstep=0.001;          // Freq step MHz
float vd1=0.2;              // Oscillo Scope Vertical mV/div
float vd2=0.2;              // Oscillo Scope Vertical mV/div
float vd3=0.2;              // Oscillo Scope Vertical mV/div
float vo1=0;                // Oscillo Scope Offset V
float vo2=0;                // Oscillo Scope Offset V
float vo3=0;                // Oscillo Scope Offset V
float sqv=-100;             // Squelch Level Value
float fhome;                // Home Freq for SCAN
float zoom;                 // Zoom rate of Screen Speana
float zosc;                 // Zoom rate of Screen Oscillo

int sd=10;                  // Speana Level dB/div
int sr=-50;                 // Speana Ref dBm
int nwsw=0;                 // Narrow 10k:0 to WIde 25M:8
int ofc=1;                  // Offset CAL ON=1/OFF=0
int mode=0;                 // Mode FM:0 AM:1
int emp=0;                  // Emphasis for NFM
int aven=10;                // Average N
int scansw=0;               // Scan mode SW
int spa=0;                  // Speana mode SW
int dem=0;                  // Demod SW
int rcd=0;                  // Record SW
int afs=0;                  // CH3 input 0:OFF 1:AF 2:POW
int xys=0;                  // XY display SW
int bws=10;                 // IIR BW kHz
int bwf=16;                 // FIR BW kHz
int ifg=0;                  // IF Gain
int att=0;                  // ATT 0:0dB 1:12dB

int sw_start=1;             // Top SW START/STOP
int sw_memin=0;             // Top SW MEMIN

int func=0;

//　IIR Coef
double a0[6], a1[6], a2[6], a3[6], a4[6];
double b0[6], b1[6], b2[6], b3[6], b4[6];

//  Xwindow
Display*       dpy;                 // Display
Pixmap         pixgrid;             // Pixmap for Grid line
Window         root;                // Root Window
Window         win;                 // Display Window
Window         wins;                // Display Window screen
Window         winp;                // Display Window parameter
Window         wint;                // Display Window top
unsigned long  black,white;         // Pixcel value
GC             gc, g_gc;            // Graphic Context
XEvent         evt;                 // Event strunc
Colormap       cmap;                // Color map
XColor         color, exact;
unsigned long  green, red, gray;
unsigned long  aqua, navy, fuchsia;
int            screen;

void GPIO_init();
void AD9218_wide(double* i, double* q);
void AD9218_narrow(double* i, double* q);
void ADF4351();
void ADF4351Q();
void FFT(double* real, double* imag, int length, int winsw);
void gpio_init();
void gpio_configure_pull (int pin, int pullmode);
void gpio_configure(int pin, int mode);
void gpio_set(int pin);
void gpio_clear(int pin);
int  gpio_read (int pin);
double fir1(double* coef, double i, int length);
double fir2(double* coef, double i, int length);
void iirlpf2coef(int n, int fc, int smplrate);
void iirhpf2coef(int n, int fc, int smplrate);
void iirlpf1coef(int n, int fc, int smplrate);
float iir(int n, float vin);
void Fircoef(double* fcoef, int fc, int length);
void Hilcoef(double* fcoef, int fc, int length);
void SetTitle(Display *display, Window window, char *title);
void DispControl();
void nw(int sw);
unsigned long RGB(Display *display,unsigned char r,unsigned char g,unsigned char b);
void getFresp(double* fresp);
void cicFresp(double* cicw, double* cicn);
void pixwrite(int W, int H, int sd, int sr, int spa, float SQL);
void TopSW();
void savemem(int ch);
void loadmem(int ch);
void copy(double* av, double* ch1, double* ch2, double* ch3, int scope_h);
void attset(int sw);

int main( void )
{
    KeySym keysym;

    FILE *fp;                       //  data out file pointer
    
    struct timeval ts, tc;          // Timer
    
    char sdata[128];
    char txt[55];

    double fresp[2200];             // freq resp comp data
    double cicw[1024];              // CIC freq resp comp data Wide
    double cicn[ 512];              // CIC freq resp comp data Narrow
    double coef1[128];              // FIR coef
    double coef2[128];              // FIR coef
    double re[1024];                // raw data I
    double im[1024];                // raw data Q
    double c3[1024];                // raw data CH3
    double db[1024];                // raw data dBm

    double av[1024];                // COPY pixcel dBm
    double ch1[1024];               // COPY pixcel CH1
    double ch2[1024];               // COPY pixcel CH2
    double ch3[1024];               // COPY pixcel CH3
    
    int n;
    int m;
    int ly;
    int li;
    int lp;
    int xsize;
    int xstart;
    int xc;                         // Speana Screen Current X
    int xl;                         // Speana Screen Last X
    int yc;                         // Speana Screen Center Y
    float xstep;
    float bzero=1;
    int tt;                         // Time for REC
    
    double am;                      // Software AGC
    double siglev;                  // Software AGC
    double sigagc;                  // Software AGC
    double afgain=1;                // SQL
    double pkdbm;                   // Peak power in FFT
    double pkfrq;                   // Peak frequency
    int scancount;                  // Counter for SACN wait
    int mkrsq;                      // Yaxis of SQL Marker Line
    double afave;                   // AF Average Power
    float vf=0;                     // Numeric KEY Resister
    float afpow;                    // AF power for NQS measure
    float cov;                      // Convert rate for Oscillo amplitude
    
    float fm;                       // FM current angle
    float lfm;                      // FM last angle
    float diff;                     // FM differential
    float sout;                     // Demod Sound Output
    float rfp;                      // RF power
    double resi;                    // FIR result I
    double resq;                    // FIR result Q
    int resw=1;                     // Rotary Encoder Event
    int lastmsb=0;                  // Narrow Buffer MSB
    int winsw=1;                    // Window SW=0:Hann 1:Blackmann-Harris
    int pns=0;                      // PN ON/OFF
    
    float offi;                     // Average offset Ich
    float offq;                     // Average offset Qch
    float offv;                     // Agerage fc
    
    int rx_r, ry_r, wx_r, wy_r;     // Mouse
    unsigned int mask_r;            // Mouse
    Window root_r, child_r;         // Mouse
    int rx_c, ry_c, wx_c, wy_c;     // Mouse
    unsigned int mask_c;            // Mouse
    Window root_c, child_c;         // Mouse
    
    x=0;                            // Demod/Display control SW
    
    int top_w;                      // Sub Window
    int top_h;                      // Sub Window
    int top_x;                      // Sub Window
    int top_y;                      // Sub Window
    int scope_w;                    // Sub Window
    int scope_h;                    // Sub Window
    int scope_x;                    // Sub Window
    int scope_y;                    // Sub Window
    int param_w;                    // Sub Window
    int param_h;                    // Sub Window
    int param_x;                    // Sub Window
    int param_y;                    // Sub Window
    
    int mousex;
    int mousey;
    
    // ALSA サウンドデバイス設定
    const static char device[] = "hw:0";
    const static snd_pcm_format_t format = SND_PCM_FORMAT_S16_LE;
    const static snd_pcm_access_t access = SND_PCM_ACCESS_RW_INTERLEAVED;
    const static unsigned int sampling_rate = 48000;    // Sampling rate = 48kHz
    const static unsigned int channels = 1;             // Mono
    const static unsigned int soft_resample = 1;        // Re-sample authorized
    const static unsigned int latency = 100000;         // 100000us
    const unsigned int buffer_size = 512;
    
    snd_pcm_t *pcm;
    int16_t buffer[ buffer_size ];

    // GPIO 初期化
    GPIO_init();
    
    // XWindow 初期化
    dpy = XOpenDisplay( "" );
    root   = DefaultRootWindow( dpy );
    screen = DefaultScreen( dpy );
    cmap   = DefaultColormap( dpy, screen );

    // 白、黒のピクセル値
    white  = WhitePixel( dpy, screen );
    black  = BlackPixel( dpy, screen );

    // カラーピクセル値
    XAllocNamedColor( dpy, cmap, "green", &color, &exact );
    green = color.pixel;

    XAllocNamedColor( dpy, cmap, "red", &color, &exact );
    red = color.pixel;
    
    XAllocNamedColor( dpy, cmap, "aqua", &color, &exact );
    aqua = color.pixel;
    
    XAllocNamedColor( dpy, cmap, "navy", &color, &exact );
    navy = color.pixel;

    XAllocNamedColor( dpy, cmap, "fuchsia", &color, &exact );
    fuchsia = color.pixel;
    
    XAllocNamedColor( dpy, cmap, "dimgray", &color, &exact );
    gray = color.pixel;
    
 
    // トップウィンドウ作成
    win  = XCreateSimpleWindow( dpy, root, WIN_X, WIN_Y, WIN_W, WIN_H, BORDER, black, RGB(dpy, 20, 20, 20));
    
    // サブウィンドウ作成（トップスイッチ画面）
    top_w=480;
    top_h=15;
    top_x=10;
    top_y=10;
    wint = XCreateSimpleWindow( dpy, win , top_x, top_y, top_w, top_h, BORDER, black, black);
    
    // サブウィンドウ作成（パラメータ設定画面）
    param_x=10;
    param_y=WIN_H-170;
    param_w=480;
    param_h=160;
    winp = XCreateSimpleWindow( dpy, win , param_x, param_y, param_w, param_h, BORDER, black, black);
    
    // サブウィンドウ作成（波形モニタ画面）
    scope_w=WIN_W-20;
    scope_h=WIN_H-param_h-top_h-40;
    scope_x=10;
    scope_y=top_h+20;
    wins = XCreateSimpleWindow( dpy, win , scope_x, scope_y, scope_w, scope_h, BORDER, black, RGB(dpy, 10, 50, 50));
    
	// GCを生成
    gc = XCreateGC( dpy, win, 0, NULL );
 
    // イベントを受け取るマスクを指定（クリックできる画面の指定）
    XSelectInput( dpy, win,  KeyPressMask | ExposureMask | ButtonPressMask);

   	// タイトルを設定
   	SetTitle(dpy, win, title);

	// サブウィンドウの可視化
    XMapSubwindows( dpy, win );
    
    // トップウィンドウの可視化
    XMapWindow( dpy, win );
    
	// Initial Setup
    GPIO_init();                    // GPIO initial
    ADF4351();                      // PLL initial
    nw(5);                          // BW=1.56M
    nwsw=5;                         // BW=1.56M
    dem=0;                          // Demod OFF
    spa=1;                          // Speana mode
    iirlpf2coef(0,  5000, 48000);   // IIR LPF default fc=10kHz
    iirlpf2coef(1,  5000, 48000);   // IIR LPF default fc=10kHz
    iirlpf1coef(2,   300, 48000);   // Emphasis fc=300Hz -6dB/oct
    iirhpf2coef(3,    50, 48000);   // DC cut fc=50Hz
    iirhpf2coef(4,   100, 48000);   // DC cut fc=100Hz
    iirhpf2coef(5,   120, 48000);   // IIR HPF fc=120Hz
    Fircoef(coef1, 10, 64);         // FIR LPF default 48kHz x 10/32 = 15kHz
    Fircoef(coef2, 10, 64);         // FIR LPF default 48kHz x 10/32 = 15kHz

    //　f特補正データの読み込み
    getFresp(fresp);
    
    //　CIC補正データの読み込み
    cicFresp(cicw, cicn);
    
    // Main Loop
    while( 1 ) {
        while ( XPending( dpy ) ) {
            
            //　イベントがあった場合の処理
            XNextEvent( dpy, &evt );

            switch( evt.type ) {
                    
                //　立ち上げ時、リサイズ時の処理
                case Expose:
                    if ( evt.xexpose.count == 0 ) {

 						Window root;
						int wx, wy;
						unsigned int width, height;
						unsigned int border;
						unsigned int depth;
				
						XGetGeometry(dpy, win, &root, &wx, &wy, &width, &height, &border, &depth);
				
						WIN_W=width;
						WIN_H=height;
                        
                        //　リサイズ時に変更されるパラメータを設定
                        param_y=WIN_H-170;
                        scope_w=WIN_W-20;
                        scope_h=WIN_H-param_h-top_h-40;

                        // 画面の移動とリサイズ
                        XMoveResizeWindow(dpy, winp, param_x, param_y, param_w, param_h);
                        XMoveResizeWindow(dpy, wins, scope_x, scope_y, scope_w, scope_h);
                        
                        // リサイズ前に古いピクスマップを破棄
                        //XFreePixmap(dpy, pixgrid);
                        
                        // リサイズに伴って新しいピクスマップを生成
                        pixgrid = XCreatePixmap(dpy, wins, scope_w, scope_h, DefaultDepth(dpy, 0));
                        g_gc = XCreateGC(dpy, pixgrid, 0, 0);
                        
                        // ウィンドウにピクスマップを設定する
                        XSetWindowBackgroundPixmap(dpy, wins, pixgrid);
                        
                        // ピクスマップにグリッドライン描画
                        pixwrite(scope_w, scope_h, sd, sr, spa, sqv);
                        
                        //　トップスイッチの表示
                        TopSW();
                        
                        //　パラメータの表示
						DispControl();
                    }
                    break;
        
                case ButtonPress:
                    // 画面上をマウスでクリックした時の処理
                    if (evt.xbutton.button==1) {                    // Mouse Left Button
                        XQueryPointer(dpy, win,
                                      &root, &child_r,
                                      &rx_r, &ry_r, &wx_r, &wy_r,
                                      &mask_r);
                        
                        //-----------------------------------------------------------------------
                        //　                 トップスイッチをクリックしたらここを実行
                        //-----------------------------------------------------------------------
                        if (wy_r>top_y && wy_r<top_y+top_h) {
                            mousex=int((wx_r-top_x)/(top_w/10));
                            if (mousex==0) sw_start=1;                  // START
                            if (mousex==1) sw_start=0;                  // STOP
                            if (mousex==2 && dem==0) {                  // DEMOD
                                dem=1;              // Demod ON
                                spa=0;              // Speana OFF
                                nwsw=2;             // BW=48k
                                nw(2);              // GPIO SET BW=48k
                                snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );   // ALSA open
                                snd_pcm_set_params( pcm, format, access, channels,          // ALSA open
                                                   sampling_rate, soft_resample, latency);  // ALSA open
                            }else if (mousex==2 && dem==1) {
                                dem=0;              // Demod OFF
                                spa=1;              // Speana ON
                                nwsw=5;             // BW=1.56M
                                nw(5);              // GPIO SET BW=1.56M
                                snd_pcm_close(pcm); // ALSA close
                            }
                            if (mousex==3 && sw_memin==1) {             // MEMIN
                                sw_memin=0;         // Clear when pushed twice
                            }else if (mousex==3 && sw_memin==0) {
                                sw_memin=1;         // MEMIN ON
                            }
                            if (mousex==4 && sw_memin==1) savemem(1);   // MEM1 SAVE
                            if (mousex==5 && sw_memin==1) savemem(2);   // MEM2 SAVE
                            if (mousex==6 && sw_memin==1) savemem(3);   // MEM3 SAVE
                            if (mousex>=4 && mousex<=6 && sw_memin==0 && dem==1){// Close ALSA before MEM Load
                                dem=0;              // Demod OFF
                                spa=1;              // Speana ON
                                nwsw=5;             // BW=1.56M
                                nw(5);              // GPIO SET BW=1.56M
                                snd_pcm_close(pcm); // ALSA close
                            }
                            if (mousex==4 && sw_memin==0) loadmem(1);   // MEM1 LOAD
                            if (mousex==5 && sw_memin==0) loadmem(2);   // MEM2 LOAD
                            if (mousex==6 && sw_memin==0) loadmem(3);   // MEM3 LOAD
                            if (mousex==7) copy(av, ch1, ch2, ch3, scope_h);                // COPY Waveform
                            if (mousex==8 && rcd==0 && sw_start==1) {   // REC Waveform
                                rcd=1;
                                gettimeofday(&ts, NULL);                                    // Get start time
                                if ((fp=fopen("recorddata.txt", "w")) == NULL) {            // Write data
                                    printf("file open error! \n");
                                }
                            }else if (mousex==8 && rcd==0 && sw_start==0) {
                                rcd=1;
                                if ((fp=fopen("recorddata.txt", "r")) == NULL) {            // Read data
                                    printf("file open error! \n");
                                }
                                printf("file open to read! \n");
                            }else if (mousex==8 && rcd==1) {                                // Close file
                                rcd=0;
                                fclose(fp);
                            }
                            if (mousex==9) {                            // Shutdown
                                XFreeGC( dpy, gc );
                                XDestroyWindow( dpy, win );
                                XCloseDisplay( dpy );
                                if (dem==1) snd_pcm_close( pcm );
                                return 0;
                            }
                            ADF4351();          // PLL Freq Set
                            ADF4351Q();         // PLL Freq Set
                            Fircoef(coef1, int(64*bwf/96), 64);
                            Hilcoef(coef2, int(64*bwf/96), 64);
                            iirlpf2coef(0, bws*500, 48000);
                            iirlpf2coef(1, bws*500, 48000);
                        }
                        
                        //-----------------------------------------------------------------------
                        //　                 スペクトラムをクリックしたらここを実行
                        //-----------------------------------------------------------------------
                        if (spa==1 && wy_r<param_y && wy_r>scope_y) {
                            if (nwsw==0) fpll=fpll+0.01000*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==1) fpll=fpll+0.02500*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==2) fpll=fpll+0.05000*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==3) fpll=fpll+0.20000*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==4) fpll=fpll+0.50000*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==5) fpll=fpll+1.00000*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==6) fpll=fpll+5.00000*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==7) fpll=fpll+10.0000*(wx_r-10-scope_w/2)/scope_w;
                            if (nwsw==8) fpll=fpll+25.0000*(wx_r-10-scope_w/2)/scope_w;
                            ADF4351();
                            func=0;
                            DispControl();
                        }
                        
                        //-----------------------------------------------------------------------
                        //　                 パラメータをクリックしたらここを実行
                        //-----------------------------------------------------------------------
                        if (wy_r>param_y) {
                            mousex=int((wx_r-param_x)/170);
                            mousey=int((wy_r-param_y-10)/17);
                            if (mousex==0 && mousey==0 && func==0) {        // Click over twice on Freq
                                mousex=int((wx_r-param_x-62)/6);
                                if (mousex==0) fstep=1000     ;
                                if (mousex==1) fstep= 100     ;
                                if (mousex==2) fstep=  10     ;
                                if (mousex==3) fstep=   1     ;
                                if (mousex==5) fstep=   0.1   ;
                                if (mousex==6) fstep=   0.01  ;
                                if (mousex==7) fstep=   0.001 ;
                                if (mousex==8) fstep=   0.0001;
                                mousex=9;       // To exit
                            }
                            if (mousex==0 && mousey==0) func=0;
                            if (mousex==0 && mousey==1) func=1;
                            if (mousex==0 && mousey==2) func=2;
                            if (mousex==0 && mousey==3) func=3;
                            if (mousex==0 && mousey==4) func=4;
                            if (mousex==0 && mousey==5) func=5;
                            if (mousex==0 && mousey==6) func=6;
                            if (mousex==0 && mousey==7) func=7;
                            if (mousex==1 && mousey==0) func=10;
                            if (mousex==1 && mousey==1) func=11;
                            if (mousex==1 && mousey==2) func=12;
                            if (mousex==1 && mousey==3) func=13;
                            if (mousex==1 && mousey==4) func=14;
                            if (mousex==1 && mousey==5) func=15;
                            if (mousex==1 && mousey==6) func=16;
                            if (mousex==2 && mousey==0) func=20;
                            if (mousex==2 && mousey==1) func=21;
                            if (mousex==2 && mousey==2) func=22;
                            if (mousex==2 && mousey==3) func=23;
                            if (mousex==2 && mousey==4) func=24;
                            if (mousex==2 && mousey==5) func=25;
                            if (mousex==2 && mousey==6) func=26;
                            if (mousex==2 && mousey==7) func=27;
                            //printf( "Click X=%4d y=%4d func=%d\n",wx_r-param_x ,wy_r-param_y,func );
                       }
                    }
                    
                    //-----------------------------------------------------------------------
                    //　                         マウス右クリックで即受信
                    //-----------------------------------------------------------------------
                    if (evt.xbutton.button==3) {                    // Mouse Right Button
                        XQueryPointer(dpy, win,
                                      &root, &child_r,
                                      &rx_r, &ry_r, &wx_r, &wy_r,
                                      &mask_r);
                    
                        // Speana/Demod Quick SW
                        if (dem==0 && wy_r<param_y) {
                            dem=1;          // Demod ON
                            spa=0;          // Speana OFF
                            nwsw=2;         // BW=48k
                            nw(2);          // GPIO SET BW=48k
                            snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );   // ALSA open
                            snd_pcm_set_params( pcm, format, access, channels,          // ALSA open
                                               sampling_rate, soft_resample, latency);  // ALSA open
                        }else if (wy_r<param_y){
                            dem=0;          // Demod OFF
                            spa=1;          // Speana ON
                            nwsw=5;         // BW=1.56M
                            nw(5);          // GPIO SET BW=1.56M
                            snd_pcm_close(pcm);                                         // ALSA close
                        }
                    }
                    
                    //-----------------------------------------------------------------------
                    //　                         マウスホイールの操作
                    //-----------------------------------------------------------------------
                    if (func== 0 && evt.xbutton.button==4 && fpll< 2200) fpll=fpll+fstep;
                    if (func== 0 && evt.xbutton.button==5 && fpll>   50) fpll=fpll-fstep;
                    if (func== 0) ADF4351Q();
                    if (func== 0 && fstep>=0) ADF4351();
                    if (func== 1 && evt.xbutton.button==4 && sd<  20   ) sd++;
                    if (func== 1 && evt.xbutton.button==5 && sd>   1   ) sd--;
                    if (func== 2 && evt.xbutton.button==4 && vd1< 10   ) vd1=vd1+0.01;
                    if (func== 2 && evt.xbutton.button==5 && vd1>  0.01) vd1=vd1-0.01;
                    if (func== 3 && evt.xbutton.button==4 && vd2< 10   ) vd2=vd2+0.01;
                    if (func== 3 && evt.xbutton.button==5 && vd2>  0.01) vd2=vd2-0.01;
                    if (func== 4 && evt.xbutton.button==4 && vd3< 10   ) vd3=vd3+0.01;
                    if (func== 4 && evt.xbutton.button==5 && vd3>  0.01) vd3=vd3-0.01;
                    if (func== 5 && evt.xbutton.button==4 && bwf< 48   ) bwf++;
                    if (func== 5 && evt.xbutton.button==5 && bwf>  2   ) bwf--;
                    if (func== 5) {
                        Fircoef(coef1, int(64*bwf/96), 64);
                        Hilcoef(coef2, int(64*bwf/96), 64);
                    }
                    if (func== 6 && evt.xbutton.button==4 && bws< 48   ) bws++;
                    if (func== 6 && evt.xbutton.button==5 && bws>  1   ) bws--;
                    if (func== 6) {
                        iirlpf2coef(0, bws*500, 48000);
                        iirlpf2coef(1, bws*500, 48000);
                    }
                    if (func== 7 && evt.xbutton.button==4 && ifg< 50   ) ifg++;
                    if (func== 7 && evt.xbutton.button==5 && ifg>  0   ) ifg--;
                    if (func==10 && evt.xbutton.button==4 && nwsw< 8   ) nwsw++;
                    if (func==10 && evt.xbutton.button==5 && nwsw> 0   ) nwsw--;
                    if (func==10) nw(nwsw);
                    if (func==10 && nwsw> 2 && dem==1) {                                // ALSA close
                        snd_pcm_close(pcm);
                        dem=0;
                    }
                    if (func==10 && nwsw<=2 && dem==0) {                                // ALSA open
                        snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );
                        snd_pcm_set_params(
                                           pcm, format, access,
                                           channels, sampling_rate, soft_resample, latency
                                           );
                        dem=1;
                    }
                    if (func==11 && evt.xbutton.button==4 && sr<    0  ) sr=sr+1;
                    if (func==11 && evt.xbutton.button==5 && sr>  -90  ) sr=sr-1;
                    if (func==12 && evt.xbutton.button==4 && vo1<  10  ) vo1=vo1+0.1;
                    if (func==12 && evt.xbutton.button==5 && vo1> -10  ) vo1=vo1-0.1;
                    if (func==13 && evt.xbutton.button==4 && vo2<  10  ) vo2=vo2+0.1;
                    if (func==13 && evt.xbutton.button==5 && vo2> -10  ) vo2=vo2-0.1;
                    if (func==14 && evt.xbutton.button==4 && vo3<  10  ) vo3=vo3+0.1;
                    if (func==14 && evt.xbutton.button==5 && vo3> -10  ) vo3=vo3-0.1;
                    if (func==15 && evt.xbutton.button==4 && sqv< -60  ) sqv++;
                    if (func==15 && evt.xbutton.button==5 && sqv>-130  ) sqv--;
                    if (func==16 && evt.xbutton.button==4 && aven<100  ) aven++;
                    if (func==16 && evt.xbutton.button==5 && aven>  1  ) aven--;
                    if (func==20 && evt.xbutton.button==4 && spa==  0  ) spa=1;
                    if (func==20 && evt.xbutton.button==5 && spa==  1  ) spa=0;
                    if (func==20 && spa==1 && dem==1) {                                 // ALSA close
                        snd_pcm_close(pcm);
                        dem=0;
                    }
                    if (func==14 && spa==0 && dem==0) {                                 // ALSA open
                        snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );
                        snd_pcm_set_params(
                                           pcm, format, access,
                                           channels, sampling_rate, soft_resample, latency
                                           );
                        dem=1;
                    }
                    if (func==21 && evt.xbutton.button==4 && ofc==  0  ) ofc=1;
                    if (func==21 && evt.xbutton.button==5 && ofc==  1  ) ofc=0;
                    if (func==22 && evt.xbutton.button==4 && emp==  0  ) emp=1;
                    if (func==22 && evt.xbutton.button==5 && emp==  1  ) emp=0;
                    if (func==23 && evt.xbutton.button==4 && xys==  0  ) xys=1;
                    if (func==23 && evt.xbutton.button==5 && xys==  1  ) xys=0;
                    if (func==24 && evt.xbutton.button==4 && mode<  2  ) mode++;
                    if (func==24 && evt.xbutton.button==5 && mode>  0  ) mode--;
                    if (func==25 && evt.xbutton.button==4 && afs<   2  ) afs++;
                    if (func==25 && evt.xbutton.button==5 && afs>   0  ) afs--;
                    if (func==26 && evt.xbutton.button==4 && scansw<2  ) scansw++;
                    if (func==26 && evt.xbutton.button==5 && scansw>0  ) scansw--;
                    if (func==26 && scansw>0) fhome=fpll;
                    if (func==27 && evt.xbutton.button==4 && att==  0  ) att=1;
                    if (func==27 && evt.xbutton.button==5 && att==  1  ) att=0;
                    if (func==27) attset(att);
                    
					DispControl();
                    TopSW();
                    pixwrite(scope_w, scope_h, sd, sr, spa, sqv);
                    break;
                    
                //-----------------------------------------------------------------------
                //　                         キー入力のイベント処理
                //-----------------------------------------------------------------------
                case KeyPress:
                    keysym = XkbKeycodeToKeysym(dpy, evt.xkey.keycode, 0, 0);
                    switch(keysym){
                        case XK_Escape:
                            // リソースの解放
                            XFreeGC( dpy, gc );
                            XDestroyWindow( dpy, win );
                            XCloseDisplay( dpy );
                            if (dem==1) snd_pcm_close( pcm );
                            return 0;
                            
                        case XK_0:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+0;
                            }else{
                                vf=vf+0*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_1:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+1;
                            }else{
                                vf=vf+1*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_2:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+2;
                            }else{
                                vf=vf+2*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_3:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+3;
                            }else{
                                vf=vf+3*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_4:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+4;
                            }else{
                                vf=vf+4*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_5:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+5;
                            }else{
                                vf=vf+5*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_6:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+6;
                            }else{
                                vf=vf+6*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_7:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+7;
                            }else{
                                vf=vf+7*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_8:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+8;
                            }else{
                                vf=vf+8*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_9:
                            // Enter Numeric
                            if (bzero==1){
                                vf=vf*10+9;
                            }else{
                                vf=vf+9*bzero;
                                bzero=bzero/10;
                            }
                            break;
                            
                        case XK_Return:
                            // 周波数決定
                            if (vf>49){
                                fpll=vf;
                                ADF4351();
                                ADF4351Q();
                                vf=0;
                                bzero=1;
                            }else{
                                vf=0;
                            }
                            break;
                            
                        case XK_period:
                            // Enter Below Zero
                            bzero=0.1;
                            break;
                            
                        case XK_F1:
                            //  BW=5kHz
                            bws=5;
                            iirlpf2coef(0, 2500, 48000);
                            iirlpf2coef(1, 2500, 48000);
                            break;
                            
                        case XK_F2:
                            //  BW=10kHz
                            bws=10;
                            iirlpf2coef(0, 5000, 48000);
                            iirlpf2coef(1, 5000, 48000);
                            break;
                            
                        case XK_F3:
                            //  BW=20kHz
                            bws=20;
                            iirlpf2coef(0, 10000, 48000);
                            iirlpf2coef(1, 10000, 48000);
                            break;
                            
                        case XK_F4:
                            //  BW=30kHz
                            bws=30;
                            iirlpf2coef(0, 15000, 48000);
                            iirlpf2coef(1, 15000, 48000);
                            break;
                            
                        case XK_F5:
                            //  BW=40kHz
                            bws=40;
                            iirlpf2coef(0, 20000, 48000);
                            iirlpf2coef(1, 20000, 48000);
                            break;
                            
                        case XK_F6:
                            //  BW=48kHz
                            bws=48;
                            break;
                            
                        case XK_F7:
                            //  BW=5kHz
                            bwf=5;
                            Fircoef(coef1, 3, 64);          //  48kHz x 3/32 = 4.5kHz
                            Fircoef(coef2, 3, 64);          //  48kHz x 3/32 = 4.5kHz
                            break;
                            
                        case XK_F8:
                            //  BW=10kHz
                            bwf=10;
                            Fircoef(coef1, 7, 64);          //  48kHz x 7/32 = 10.5kHz
                            Fircoef(coef2, 7, 64);          //  48kHz x 7/32 = 10.5kHz
                            break;
                            
                        case XK_F9:
                            //  BW=20kHz
                            bwf=20;
                            Fircoef(coef1, 13, 64);         //  48kHz x 13/32 = 19.5kHz
                            Fircoef(coef2, 13, 64);         //  48kHz x 13/32 = 19.5kHz
                            break;
                            
                        case XK_F10:
                            //  BW=30kHz
                            bwf=30;
                            Fircoef(coef1, 20, 64);         //  48kHz x 20/32 = 30.0kHz
                            Fircoef(coef2, 20, 64);         //  48kHz x 20/32 = 30.0kHz
                            break;
                            
                        case XK_F11:
                            //  BW=40kHz
                            bwf=40;
                            Fircoef(coef1, 27, 64);         //  48kHz x 27/32 = 40.5kHz
                            Fircoef(coef2, 27, 64);         //  48kHz x 27/32 = 40.5kHz
                            break;
                            
                        case XK_F12:
                            //  BW=48kHz
                            bwf=48;
                            Fircoef(coef1, 32, 64);         //  48kHz x 32/32 = 48.0kHz
                            Fircoef(coef2, 32, 64);         //  48kHz x 32/32 = 48.0kHz
                            break;
                            
						case XK_Page_Down:
							//  Function select down
							if (func<30) func++;
							break;

						case XK_Page_Up:
							//  Function select up
							if (func>0) func--;
							break;

                        case XK_Left:
                            if (func==0 && fstep<100 ) {fstep=fstep*10;}    //  Freq step up
                            break;
                            
                        case XK_Right:
                            if (func==0 && fstep>0.0001) {fstep=fstep/10;}    //  Freq step down
                            break;
                            
                        case XK_Up:
							if (func==0) {
                            	fpll=fpll+fstep;
                            	ADF4351(); }
                            break;
                            
                        case XK_Down:
                            if (func==0) {
                            	fpll=fpll-fstep;
                            	ADF4351(); }
                            break;
                            
                        case XK_space:
                            // Speana/Demod Quick SW
                            if (dem==0) {
                                dem=1;          // Demod ON
                                spa=0;          // Speana OFF
                                nwsw=2;         // BW=48k
                                nw(2);          // GPIO SET BW=48k
                                snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );   // ALSA open
                                snd_pcm_set_params( pcm, format, access, channels,          // ALSA open
                                                   sampling_rate, soft_resample, latency);  // ALSA open
                            }else{
                                dem=0;          // Demod OFF
                                spa=1;          // Speana ON
                                nwsw=5;         // BW=1.56M
                                nw(5);          // GPIO SET BW=1.56M
                                snd_pcm_close(pcm);                             // ALSA close
                            }
                            break;
                            
                        case XK_p:
                            // PN off TX RX
                            pns++;
                            if (pns>2) pns=0;
                            if (pns==0) {
                                gpio_clear(20);
                                gpio_clear(26);
                                printf("PN off \n");
                            }
                            if (pns==1) {
                                gpio_clear(20);
                                gpio_set(26);
                                printf("PN TX on \n");
                            }
                            if (pns==2) {
                                gpio_set(20);
                                gpio_set(26);
                                printf("PN TX & RX on \n");
                            }
                            break;
                            
                        case XK_a:
                            // AF sound
                            if (afs==0) {
                                afs=1;
                            }else{
                                afs=0;
                            }
                            break;
                            
                        case XK_w:
                            // FFT Window select
                            if (winsw==0) {
                                winsw=1;
                                printf("Blackman-Harris window \n");
                            }else if(winsw==1) {
                                winsw=2;
                                printf("Square window \n");
                            }else if(winsw==2) {
                                winsw=0;
                                printf("Hann window \n");
                            }
                            break;
                            
                        case XK_Tab:
                            //　オシロ、スペアナ切り替え
                            if (spa==0) {
                                spa=1;
                            }else{
                                spa=0;
                            }
                            break;
                            
                        case XK_r:
                            //  Record start or stop
                            if (rcd==0 && sw_start==1) {
                                rcd=1;
                                gettimeofday(&ts, NULL);    // Get start time
                                if ((fp=fopen("recorddata.txt", "w")) == NULL) {
                                    printf("file open error! \n");
                                }
                            }else if (rcd==0 && sw_start==0) {
                                rcd=1;
                                if ((fp=fopen("recorddata.txt", "r")) == NULL) {
                                    printf("file open error! \n");
                                }
                                printf("file open to read! \n");
                            }else if (rcd==1) {
                                rcd=0;
                                fclose(fp);
                            }
                            break;
                            
                    }
					DispControl();
                    pixwrite(scope_w, scope_h, sd, sr, spa, sqv);
             }
        }
        
        //　ロータリエンコーダ処理
        if (gpio_read(13)==1 && resw==0 && pns==0)
        {
            fpll=fpll-fstep;
            resw=1;
            ADF4351Q();
 			DispControl();
        }
        if (gpio_read(19)==1 && resw==0 && pns==0)
        {
            fpll=fpll+fstep;
            resw=1;
            ADF4351Q();
			DispControl();
        }
        if (gpio_read(13)==0 && gpio_read(19)==0) {
            resw=0;
        }
        
        //  *******************************************************************************
        //　                                   スキャンモード
        //  *******************************************************************************
        if (scansw>0) {
            //　スペアナ監視モード時にスケルチレベルを超える信号に同調、受信モードに移行
            if (spa==1 && pkdbm>sqv) scancount++;
            if (spa==1 && pkdbm<sqv) scancount=0;
            if (spa==1 && scancount>aven) {     // 30ms*aven(10)=300ms
                printf("Scan Detect Freq=%8.3f MHz Level=%5.1f dBm\n",pkfrq,pkdbm);
                fpll=pkfrq;     // Tuning Target Freq
                ADF4351Q();     // ADF4351 Quick Tune
                scancount=0;    // Reset count
                dem=1;          // Demod ON
                spa=0;          // Speana OFF
                nwsw=2;         // BW=48k
                nw(2);          // GPIO SET BW=48k
                snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );   // ALSA open
                snd_pcm_set_params( pcm, format, access, channels,          // ALSA open
                                   sampling_rate, soft_resample, latency);  // ALSA open
                if (scansw==2) scansw=0;                        // Quit SCAN when SNGL mode
                DispControl();                                  // Parameter Update
                TopSW();                                        // Top SW Update
                pixwrite(scope_w, scope_h, sd, sr, spa, sqv);   // Pixmap rewrite
             }
            //　受信動作時にスケルチレベルを信号が下回ったらスペアナ監視モードに移行
            if (spa==0 && sigagc<sqv) scancount++;
            if (spa==0 && sigagc>sqv) scancount=0;
            if (spa==0 && scancount>aven*10){    // 10ms*aven(10)*10=1000ms
                fpll=fhome;     // Return to Home Freq
                ADF4351Q();     // ADF4351 Quick Tune
                scancount=0;    // Reset count
                dem=0;          // Demod OFF
                spa=1;          // Speana ON
                nwsw=5;         // BW=1.56M
                nw(5);          // GPIO SET BW=1.56M
                snd_pcm_close(pcm);                             // ALSA close
                DispControl();                                  // Parameter Update
                TopSW();                                        // Top SW Update
                pixwrite(scope_w, scope_h, sd, sr, spa, sqv);   // Pixmap rewrite
            }
        }
        
        //  *******************************************************************************
        //　                            波形表示＆復調タイミング制御
        //  *******************************************************************************
        if (nwsw>2) {
            //  Wide or Mid Band
            x=1;
            usleep(30000);
            AD9218_wide(re, im);
        }
        if (nwsw<=2 && spa==1) {
            //  Narrow Band スペアナでADデータ読出しとADサンプリングがぶつからないようにする
            if (gpio_read(5)==0 && lastmsb==0) {
                //　復調動作、画面表示を停止、ADCデータを1回だけ読込
                x=2;
                AD9218_narrow(re, im);
                lastmsb=1;
            }
            if (gpio_read(5)==1 && lastmsb==1) {
                //　復調動作、画面表示を１回だけ実行
                x=0;
                lastmsb=0;
            }
        }
        if (nwsw<=2 && spa==0) {
            //  Narrow Band リアルタイム復調処理
            //  MSBが変化するまで待つ
            if (gpio_read(5)==lastmsb) {
                //　復調動作を禁止、待機状態、画面表示を１回だけ実行
                x++;
            }else{
                //　復調動作を許可、MSB読み出しとADデータ取得
                x=0;
                lastmsb=gpio_read(5);
                AD9218_narrow(re, im);
            }
        }
        
        //  *******************************************************************************
        //　                                     復調処理
        //　                         画面表示と同じタイミングでは実行しない
        //  *******************************************************************************
        if (x==0 && sw_start==1) {
            
            afpow=0;
            siglev=0;
            
            for (m=0; m<512; m++) {
                
                resi=re[m]/fresp[int(fpll)];        //　f特補正　Ich
                resq=im[m]/fresp[int(fpll)];        //　f特補正　Qch
                
                // ------------------------------------------------------------------------
                // ----------------------------  IFフィルタ処理  ----------------------------
                // ------------------------------------------------------------------------
                
                resi=fir1(coef1,resi,64);
                resq=fir2(coef1,resq,64);
                
                re[n]=re[n]/fresp[int(fpll)];
                im[n]=im[n]/fresp[int(fpll)];
               
                if (bws<48) {
                    resi=iir(0, resi);              // IIR filter Ich
                    resq=iir(1, resq);              // IIR filter Qch
                }
                
                // DCオフセット除去
                offv=0.001;                         // fc = 48kHz * 0.001 = 48Hz
                offi=(1-offv)*offi+offv*resi;
                offq=(1-offv)*offq+offv*resq;
                resi=resi-offi;
                resq=resq-offq;
                
                if (offi>1 || offi<-1) offi=0;      // In case of satulate
                if (offq>1 || offq<-1) offq=0;      // In case of satulate
                
                // Software AGC
                am=sqrtf(resi*resi+resq*resq);
                siglev=siglev+am;
                
                // ------------------------------------------------------------------------
                // ----------------------------  Narrow FM 復調 ----------------------------
                // ------------------------------------------------------------------------
                if (mode==0) {
                    fm=atan2f(resi,resq);           // ArcTan -pi to pi
                    diff = fm-lfm;                  // Differential
                    if (fabs(diff)<M_PI) {
                        sout = diff;
                    }else{
                        if(diff>0) {
                            sout = diff - 2 * M_PI;
                        }else{
                            sout = diff + 2 * M_PI;
                        }
                    }
                    lfm=fm;
                    
                    if (emp==1) {                   // Emphasis mode ON
                        sout=0.2*iir(2,sout);       // Emphasis fc=300Hz 6dB/oct
                        //sout=iir(5,sout);           // HPF fc=120Hz
                        afpow=afpow+fabs(sout);
                    }
                }
                
                // ------------------------------------------------------------------------
                // -------------------------------  AM 復調 --------------------------------
                // ------------------------------------------------------------------------
                if (mode==1) {
                    am=10*am*pow(10,-(sigagc+80)/20);       // AGC ゲイン調整 -80dBmを基準
                    sout=am-0.6;                            // DCオフセット除去　ALSA不具合対策
                }
                
                // ------------------------------------------------------------------------
                // -------------------------------  DSB 復調 -------------------------------
                // ------------------------------------------------------------------------
                if (mode==2) {
                    sout=(resi+resq)*pow(10,(float)ifg/20); // IG Gain
                }
                
                buffer[m] = (int)(sout * afgain * 10000);   // ALSA sound buffer +/-3.14 x 10000
                
                re[m]=resi;
                im[m]=resq;

                //　CH3に復調出力を接続
                if (afs==1) c3[m]=sout;
            }

            // ----------------------------------------------------------------------------
            // -------------------------  ALSAバッファへ音声出力 -----------------------------
            // ----------------------------------------------------------------------------
            if (dem==1 && spa==0) {
                int write_result = snd_pcm_writei ( pcm, ( const void* )buffer, buffer_size );
                if( write_result < 0 ) {
                    if( snd_pcm_recover( pcm, write_result, 0 ) < 0 ) {
                        printf( "Unable to recover this stream." );
                        snd_pcm_close( pcm );
                        abort();
                    }
                }
            }
            
            if (spa==1) x=1;        // Remove DC and Display FFT if in Narrow Speana
        }
        
        //  *******************************************************************************
        //　                              復調ストリームデータの保存
        //　                         復調処理と同じタイミングでは実行しない
        //  *******************************************************************************
        if (x==1 && sw_start==1 && rcd==1) {
            
            //　ストリーム音声データ（途切れるため推奨しない）
            if (afs==1){
                for (m=0; m<512; m++) {
                    sprintf(sdata, "%9.6f\n", c3[m]);
                    fputs(sdata, fp);
                }
            }
            
            //　ストリーム振幅データ（単位V）
            if (afs==2){
                gettimeofday(&tc, NULL);                                            // meas curr time
                tt=(tc.tv_sec - ts.tv_sec) * 1000 + (tc.tv_usec - ts.tv_usec)/1000; // calc timer
                sprintf(sdata, " %d,%12.10f\n", tt, 0.01*cov*c3[511]);
                fputs(sdata, fp);
            }
        }
        
        //  *******************************************************************************
        //　                              復調ストリームデータの再生
        //　                             　　STOP時のみ実行可能
        //  *******************************************************************************
        if (sw_start==2 && rcd==1) {    //　sw_start==2 で使用不可　sw_start==0 で使用可能
            
            x==99;
            //usleep(8000);
            
            if (dem==0) {
                dem=1;
                snd_pcm_open( &pcm, device, SND_PCM_STREAM_PLAYBACK, 0 );   // ALSA open
                snd_pcm_set_params( pcm, format, access, channels,          // ALSA open
                                   sampling_rate, soft_resample, latency);  // ALSA open
            }
            
            for (m=0; m<512; m++) {
                fgets(sdata, sizeof(sdata), fp);
                c3[m] = atof(sdata);
                buffer[m] = (int)(c3[m] * 10000);    // ALSA sound buffer +/-3.14 x 10000
            }
            
            int write_result = snd_pcm_writei ( pcm, ( const void* )buffer, buffer_size );
            if( write_result < 0 ) {
                if( snd_pcm_recover( pcm, write_result, 0 ) < 0 ) {
                    printf( "Unable to recover this stream." );
                    snd_pcm_close( pcm );
                    abort();
                }
            }
        }
        
        //  *******************************************************************************
        //　                                     画面表示
        //　                         復調処理と同じタイミングでは実行しない
        //  *******************************************************************************
        if (x==1 && sw_start==1) {
            
            //　画面サイズの定義
            if (spa==1) {
                //　スペアナ画面設定
                if (nwsw<=2) {      // Narrow
                    xstart=0;
                    xsize=512;
                    xstep=float(scope_w)/xsize;
                }else{              // Wide
                    xstart=0;
                    xsize=1024;
                    xstep=float(scope_w)/xsize;
                }
            }else{
                //　オシロ画面設定
                xstart=0;
                xsize=512;
                xstep=(float(scope_w)/xsize)/zosc;
            }

            //　波形振幅のf特補正（BW=48kの二重補正を防止）
            if (nwsw>2) {
                for (n=xstart; n<xsize; n++) {
                    re[n]=re[n]/fresp[int(fpll)];
                    im[n]=im[n]/fresp[int(fpll)];
                }
            }
            
            //　FFT処理
            if (spa==1 && nwsw>2 ) FFT(re, im, 1024, winsw);
            if (spa==1 && nwsw<=2) FFT(re, im, 512 , winsw);
        
            // CH3電力波形バッファデータのシフト
            if (spa==0 && afs==2) {
                for (n=0; n<xsize; n++) {
                    c3[n]=c3[n+1];
                }
                c3[511]=0;
            }
            
            //　画面のクリア
            XClearWindow(dpy, wins);
            
            pkdbm=-999;
            
            for (n=0; n<xsize; n++) {
                
                // I/Q値より振幅を求める
                if (nwsw<=2          ) rfp=sqrtf((re[n]*re[n]+im[n]*im[n]))/64;       // BW=48.8kHz
                if (nwsw>2 && nwsw<=5) rfp=sqrtf((re[n]*re[n]+im[n]*im[n]))/32;       // BW=15.6MHz
                if (nwsw>5           ) rfp=sqrtf((re[n]*re[n]+im[n]*im[n]))/16;       // BW=50.0MHz
                
                // 掃引1本分の平均電力を得る
                if (spa==0 && afs==2) {
                    c3[511]=c3[511]+sqrtf(re[n]*re[n]+im[n]*im[n])/xsize;
                }
                
                //　オシロ振幅補正係数
                if (nwsw<=2          ) cov=0.05;         // BW=48.8kHz
                if (nwsw>2 && nwsw<=5) cov=0.20;         // BW=15.6MHz
                if (nwsw>5           ) cov=0.40;         // BW=50.0MHz
                
                if (spa==0) {
                    //　オシロ表示
                    if (xys==0) {
                        ch1[n]=(float)scope_h/2-cov*(re[n] * (float)scope_h/vd1) - (float)scope_h*vo1/10;
                        ch2[n]=(float)scope_h/2-cov*(im[n] * (float)scope_h/vd2) - (float)scope_h*vo2/10;
                        ch3[n]=(float)scope_h/2-cov*(c3[n] * (float)scope_h/vd3) - (float)scope_h*vo3/10;
                    }else{
                        ch1[n]=(float)scope_h/2-cov*(re[n] * (float)scope_h/vd1) - (float)scope_h*vo1/10;
                        ch2[n]=(float)scope_w/2-cov*(im[n] * (float)scope_h/vd2) + (float)scope_w*vo2/10;
                    }
                }else{
                    
                    //　スペアナ表示、受信レベルをdBmへ換算する
                    if (nwsw>2 && nwsw<=5) rfp=rfp/cicw[n];           // CIC fresp comp BW=1.56M
                    if (nwsw<=2          ) rfp=rfp/cicn[n];           // CIC fresp comp BW=48.8k
                    db[n]=20*log10f(rfp)-70;
                    
                    //  ピークサーチ
                    if (pkdbm<db[n]) {
                        pkdbm=db[n];
                        if (n<256 && nwsw<=2          ) pkfrq=fpll+      n *0.0488/xsize;
                        if (n>255 && nwsw<=2          ) pkfrq=fpll-( 512-n)*0.0488/xsize;
                        if (n<512 && nwsw>2 && nwsw<=5) pkfrq=fpll+      n *1.5625/xsize;
                        if (n>511 && nwsw>2 && nwsw<=5) pkfrq=fpll-(1024-n)*1.5625/xsize;
                        if (n<512 && nwsw>5           ) pkfrq=fpll+      n *50.000/xsize;
                        if (n>511 && nwsw>5           ) pkfrq=fpll-(1024-n)*50.000/xsize;
                    }
                    
                    //　FFT表示
                    ch1[n]=-(((float)scope_h/10.0)/sd)*db[n]+(float)sr*((float)scope_h/10.0)/(float)sd;
                    av[n]=av[n]*(1.0-1.0/(float)aven)+(float)ch1[n]/(float)aven;        // Averaging
                    if (av[n]<0 || av[n]>1000) av[n]=500;                               // In case of satulate
                    ch1[n]=av[n];
                }
                
                //　画面端処理
                if (ch1[n]<  0) ch1[n]=  0;
                if (ch1[n]>scope_h) ch1[n]=scope_h;
                if (ch2[n]<  0) ch2[n]=  0;
                //if (ch2[n]>scope_h) ch2[n]=scope_h;
                if (ch3[n]<  0) ch3[n]=  0;
                if (ch3[n]>scope_h) ch3[n]=scope_h;
                if (xys==0) {
                    if (ch2[n]>scope_h) ch2[n]=scope_h;
                }else{
                    if (ch2[n]>scope_h*2) ch2[n]=scope_h*2;
                    //if (ch2[n]>scope_h) ch2[n]=scope_h;
                }
                
                //　オシロ波形描画
                if (n>1 && spa==0) {
                    if (xys==0) {
                        XSetForeground( dpy, gc, aqua );                                            // CH1 AQUA
                        XDrawLine( dpy, wins, gc, xstart+(n-1)*xstep, ly, xstart+(n)*xstep, ch1[n]);
                        XSetForeground( dpy, gc, fuchsia  );                                        // CH2 fuchsia
                        XDrawLine( dpy, wins, gc, xstart+(n-1)*xstep, li, xstart+(n)*xstep, ch2[n]);
                        if (afs>0) {
                            XSetForeground( dpy, gc, green);                                        // CH3 GREEN
                            XDrawLine( dpy, wins, gc, xstart+(n-1)*xstep, lp, xstart+(n)*xstep, ch3[n]);
                        }
                    }else{
                        //　XY表示
                        XSetForeground( dpy, gc, aqua );                                            // XY AQUA
                        XDrawLine( dpy, wins, gc, li, ly, ch2[n], ch1[n]);
                    }
                }
                
                //　スペアナ波形表示
                XSetForeground( dpy, gc, green );                                                   // Speana green
                if (n==0) yc=ch1[n];
                if (n>0 && n<(0.5*zoom)*xsize && spa==1) {
                    xl=scope_w/2+(n-1)*xstep/zoom;
                    xc=scope_w/2+(n  )*xstep/zoom;
                    XDrawLine( dpy, wins, gc, xl, ly, xc, ch1[n]);
                }
                if (n>=(1.0-0.5*zoom)*xsize && spa==1) {
                    xl=(-(1.0-0.5*zoom)*xsize+n-1)*xstep/zoom;
                    xc=(-(1.0-0.5*zoom)*xsize+n  )*xstep/zoom;
                    XDrawLine( dpy, wins, gc, xl, ly, xc, ch1[n]);
                }
                if (n==xsize-1 && spa==1) XDrawLine( dpy, wins, gc, scope_w/2, yc, xc, ch1[n]);     // Draw Center Zero Hz
                
                ly=ch1[n];
                li=ch2[n];
                lp=ch3[n];
                
            }
            
            //  Emphasis
            if (emp==1 && spa==0)  {
                afave=0.9*afave+0.1*afpow/512;
                sprintf(txt, "NFM Empasis ON AF=%5.1f dB",20*log10f(afave));
                XSetForeground( dpy, gc, white );
                XDrawString(dpy, wins, gc, 10, 35, txt, 26);
            }
            
            //  Get SIG Level
            sigagc=20*log10f(siglev)-110;
            
            //  スケルチ閾値判定
            if (sigagc<sqv) {
                afgain=0.01;
            }else{
                afgain=1;
            }
            
            //  Display SIG Level
            if (spa==0 && nwsw<=2) {
                sprintf(txt, "SIG Level= %6.1f dBm",sigagc);
                XSetForeground( dpy, gc, aqua );
                XDrawString(dpy, wins, gc, 10, 15, txt, 21);
            }
            if (spa==1) {
                //  Speana Peak Level
                //sprintf(txt, "Peak Level= %6.1f dBm at %8.3f MHz",pkdbm ,pkfrq);
                if (nwsw>5           ) sprintf(txt, "Peak Level= %6.1f dBm at %8.3f MHz RBW= 48.8 kHz",pkdbm ,pkfrq);
                if (nwsw<=5 && nwsw>2) sprintf(txt, "Peak Level= %6.1f dBm at %8.3f MHz RBW= 1.52 kHz",pkdbm ,pkfrq);
                if (nwsw<=2          ) sprintf(txt, "Peak Level= %6.1f dBm at %8.3f MHz RBW= 95.3 Hz ",pkdbm ,pkfrq);
                XSetForeground( dpy, gc, aqua );
                XDrawString(dpy, wins, gc, 10, 15, txt, 52);
            }
        }
    }
}


//-------------------------------------------------------------------
//                          FIR Filter1
//-------------------------------------------------------------------
double fir1(double* coef, double i, int length)
{
    int k;
    double ri=0;
    
    static double z[128];
    
    for (k=0; k<length-1; k++) {
        z[length-1-k] = z[length-2-k];
        ri = ri + coef[length-1-k] * z[length-1-k];
    }
    
    z[0] = i;
    ri = ri+ coef[0] * z[0];
    
    return ri;
}

//-------------------------------------------------------------------
//                          FIR Filter2
//-------------------------------------------------------------------
double fir2(double* coef, double i, int length)
{
    int k;
    double ri=0;
    
    static double z[128];
    
    for (k=0; k<length-1; k++) {
        z[length-1-k] = z[length-2-k];
        ri = ri + coef[length-1-k] * z[length-1-k];
    }
    
    z[0] = i;
    ri = ri+ coef[0] * z[0];
    
    return ri;
}

//-------------------------------------------------------------------
//                          A/Dコンバータ受信　50Msps 1.56Msps
//-------------------------------------------------------------------
void AD9218_wide(double* i, double* q)
{
    double val;
    int v;
    int lv;
    int gv=0;
    int n;
    int m;
    
    static double avi;
    static double avq;
    static double ofi;
    static double ofq;
    float avec=0.001;
    
    gpio_set(11);       // sce on
    gpio_set(6);        // Ich select
    
    //  Ich Load
    for (m=0; m<1024; m++) {
        
        gv=0;
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read( 9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;         //  -511 to 511
        val=(double)gv/(double)511;     //  -1 to 1
        i[m]=val*0.5;                   //  -0.5 to 0.5
        
        avi=avi*(1-avec)+val*0.5*avec;
    }
    
    gpio_clear(6);       // Qch select
    
    //  Qch Load
    for (m=0; m<1024; m++) {
        
        gv=0;
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read( 9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;         //  -511 to 511
        val=(double)gv/(double)511;     //  -1 to 1
        q[m]=val*0.5;                   //  -0.5 to 0.5
        
        avq=avq*(1-avec)+val*0.5*avec;
    }
    
    if (ofc==1) {                       //  CAL Offset ON/OFF
        ofi=avi;
        ofq=avq;
    }
    
    for (m=0; m<1024; m++) {            // Remove Offset
        if (att==1 && nwsw>2 && nwsw<=5){
            i[m]=(i[m]-ofi)*4.0;
            q[m]=(q[m]-ofq)*4.0;
        }else{
            i[m]=i[m]-ofi;
            q[m]=q[m]-ofq;
        }
    }
    
    gpio_clear(11);                     //  CE0 0
    
}

//-------------------------------------------------------------------
//                          A/Dコンバータ受信　48ksps
//-------------------------------------------------------------------
void AD9218_narrow(double* i, double* q)
{
    double val;
    int v;
    int lv;
    int gv=0;
    int n;
    int m;
    
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on
    
    gpio_set(6);        // Ich select
    
    //  Ich Load
    for (m=0; m<512; m++) {
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=0;
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read( 9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;          //  Comp on 2
        val=(double)gv/(double)511;
        if (att==0) i[m]=val;
        if (att==1) i[m]=val*4;
        
    }
    
    
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off
    gpio_clear(11);     // sce off

    gpio_clear(6);      // Qch select
    
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on
    gpio_set(11);       // sce on

    
    //  Qch Load
    for (m=0; m<512; m++) {
        
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        gpio_set(7);           //  CK 1
        
        gv=0;
        gv=gv+gpio_read(14)*512;
        gv=gv+gpio_read(15)*256;
        gv=gv+gpio_read(17)*128;
        gv=gv+gpio_read(27)*64;
        gv=gv+gpio_read(22)*32;
        gv=gv+gpio_read(23)*16;
        gv=gv+gpio_read(24)*8;
        gv=gv+gpio_read(10)*4;
        gv=gv+gpio_read( 9)*2;
        gv=gv+gpio_read(25)*1;
        
        gpio_clear(7);         //  CK 0
        
        if (gv>511) gv=gv-1023;         //  Comp on 2
        val=(double)gv/(double)511;
        q[m]=val;
        if (att==0) q[m]=val;
        if (att==1) q[m]=val*4;
        
    }
    
    gpio_clear(11);                     //  sce off
    
}

//-------------------------------------------------------------------
//                        シンセサイザ　ADF4531　設定
//-------------------------------------------------------------------
void ADF4351()
{
    int n;
    int m;
    int d[6];
    int b;
    
    float fo;
    float fstp;
    int fint;
    int div;
    int frac;
    int mod;
    int mmod;
    
    int i1,i2,i3,i4,m1,m2;
    
    float floop=5.000;      // Freq loopfilter=5MHz
    
    if (fpll>2195)  fpll=2195;
    if (fpll<  50)  fpll=  50;
    
    fo=fpll*2;
    div=pow(2,int(log2(4400/fo)));
    fstp=0.0016/div;                // 0.0016MHz/8=200Hz
    fint=int(fo*div/floop);
    mod=int(floop/(fstp*div));
    frac=int(mod*(fo*div/floop-fint));
    
//        printf("\n");
//        printf("fset=%5.2f MHz\n",fpll);
//        printf("fosc=%5.2f MHz\n",fo);
//        printf("Int=%5d \n",fint);
//        printf("Div=%2d \n",div);
//        printf("Frac=%4d \n",frac);
//        printf("Mod=%4d \n",mod);
//        printf("fvco=%5.2f MHz\n",fo*div);
    
    d[0]=fint*0x8000+frac*8;
    
    d[1]=0x08008001+mod*8;
    
    d[2]=0;
    d[2]=d[2]+0x004000 * 5;             //  R counter
    d[2]=d[2]+0x000200 * 2;             //  Icp 0.31mA x N
    d[2]=d[2]+0x000040 * 1;             //  PD polarity 0:Neg 1:Pos
    d[2]=d[2]+0x000001 * 2;             //  Control bit 2
    
    d[3]=0;
    d[3]=d[3]+0x200000 * 0;             //  Charge Cancelation Disable
    d[3]=d[3]+0x040000 * 1;             //  Cycle Slip Reduction Enable
    d[3]=d[3]+0x008000 * 1;             //  Fast lock Enable
    d[3]=d[3]+0x000008 * 2000;          //  Fast lock CK divider t=2000/25M=80us
    d[3]=d[3]+0x000001 * 3;             //  Control bit 3
    
    d[4]=0;
    d[4]=d[4]+0x800000 * 1;             //  Feedback 0:divided 1:fundamental
    d[4]=d[4]+0x100000 * log2(div);     //  RF Divider /1 to /64
    d[4]=d[4]+0x001000 * 0xC8;          //  Clock Divider
    d[4]=d[4]+0x000800 * 0;             //  VCO Power Down 0:UP 1:DOWN
    d[4]=d[4]+0x000400 * 0;             //  Mute Til Lock 0:Disable 1:Enable
    d[4]=d[4]+0x000200 * 0;             //  Aux output select 0:divided 1:fundamental
    d[4]=d[4]+0x000100 * 0;             //  Aux Out 0:Disable 1:Enable
    d[4]=d[4]+0x000040 * 0;             //  Aux Out 0:-4dBm 1:-1dBm 2:+2dBm 3:+5dBm
    d[4]=d[4]+0x000020 * 1;             //  RF Out 0:Disable 1:Enable
    d[4]=d[4]+0x000008 * 3;             //  RF Out 0:-4dBm 1:-1dBm 2:+2dBm 3:+5dBm
    d[4]=d[4]+0x000001 * 4;             //  Control bit 4
    
    d[5]=0x580005;                      //  Digital Lock Detect
    
    gpio_clear(12);                // LE L
    gpio_clear(2);                 // CK L
    gpio_clear(3);                 // DATA L
    
    for (m=0; m<6; m++) {
        
        gpio_clear(12);            // LE L
        
        for (n=0; n<32; n++) {
            b=d[m] & (int)pow(2,31-n);
            if (n==0) b=0;
            
            gpio_clear(2);         // CK L
            
            if (b==0) {
                gpio_clear(3);     // DATA H
            }else{
                gpio_set(3);       // DATA L
            }
            
            gpio_set(2);           // CK H
        }
        
        gpio_clear(2);             // CK L
        gpio_clear(3);             // DATA L
        gpio_set(12);              // LE H
        
    }
    
    gpio_clear(12);                // LE L
    gpio_clear(2);                 // CK L
    gpio_clear(3);                 // DATA L
    
}

//-------------------------------------------------------------------
//                        シンセサイザ　ADF4531　設定（高速版）
//-------------------------------------------------------------------
void ADF4351Q()
{
    int n;
    int m;
    int d[6];
    int b;
    
    float fo;
    float fstp;
    int fint;
    int div;
    int frac;
    int mod;
    int mmod;
    
    static int ldiv;
    
    int i1,i2,i3,i4,m1,m2;
    
    float floop=5.000;              // Freq loopfilter=5MHz
    
    if (fpll>2195)  fpll=2195;
    if (fpll<  50)  fpll=  50;
    
    fo=fpll*2;
    div=pow(2,int(log2(4400/fo)));
    fstp=0.0016/div;                // 0.0016MHz/8=200Hz
    fint=int(fo*div/floop);
    mod=int(floop/(fstp*div));
    frac=int(mod*(fo*div/floop-fint));
    
    d[0]=fint*0x8000+frac*8;
    
    d[1]=0x08008001+mod*8;
    
    d[4]=d[4] & 0xFF8FFFFF;             //  Clear RF Divider 3bit
    d[4]=d[4]+0x100000 * log2(div);     //  RF Divider /1 to /64

    gpio_clear(12);                // LE L
    gpio_clear(2);                 // CK L
    gpio_clear(3);                 // DATA L
    
    for (m=0; m<2; m++) {
        
        gpio_clear(12);            // LE L
        
        for (n=0; n<32; n++) {
            b=d[m] & (int)pow(2,31-n);
            if (n==0) b=0;
            
            gpio_clear(2);         // CK L
            
            if (b==0) {
                gpio_clear(3);     // DATA H
            }else{
                gpio_set(3);       // DATA L
            }
            
            gpio_set(2);           // CK H
        }
        
        gpio_clear(2);             // CK L
        gpio_clear(3);             // DATA L
        gpio_set(12);              // LE H
        
    }
    
    if (ldiv!=div) {
        
        m=4;
        
        gpio_clear(12);            // LE L
        
        for (n=0; n<32; n++) {
            b=d[m] & (int)pow(2,31-n);
            if (n==0) b=0;
            
            gpio_clear(2);         // CK L
            
            if (b==0) {
                gpio_clear(3);     // DATA H
            }else{
                gpio_set(3);       // DATA L
            }
            
            gpio_set(2);           // CK H
        }
        
        gpio_clear(2);             // CK L
        gpio_clear(3);             // DATA L
        gpio_set(12);              // LE H
        
    }
    
    gpio_clear(12);                // LE L
    gpio_clear(2);                 // CK L
    gpio_clear(3);                 // DATA L
    
    ldiv=div;
}

//-------------------------------------------------------------------
//                          GPIO　初期化
//-------------------------------------------------------------------
void GPIO_init()
{
    
    gpio_init();                            //
        
    gpio_configure(2,  GPIO_OUTPUT);        //  PLL CK
    gpio_configure(3,  GPIO_OUTPUT);        //  PLL DA
    gpio_configure(4,  GPIO_INPUT);         //  dummy
    gpio_configure(5,  GPIO_INPUT);         //  msbsw
    gpio_configure(6,  GPIO_OUTPUT);        //  iqsw
    gpio_configure(7,  GPIO_OUTPUT);        //  SCLK
    gpio_configure(8,  GPIO_OUTPUT);        //  nwsw
    gpio_configure(9,  GPIO_INPUT);         //  dat1
    gpio_configure(10, GPIO_INPUT);         //  dat2
    gpio_configure(11, GPIO_OUTPUT);        //  SCE
    gpio_configure(12, GPIO_OUTPUT);        //  PLL LE
    gpio_configure(13, GPIO_INPUT);         //  Rotary Encoder A
    gpio_configure(14, GPIO_INPUT);         //  dat9
    gpio_configure(15, GPIO_INPUT);         //  dat8
    gpio_configure(16, GPIO_OUTPUT);        //  Mid/Wide Band Select
    gpio_configure(17, GPIO_INPUT);         //  dat7
    gpio_configure(18, GPIO_INPUT);         //  pwm
    gpio_configure(19, GPIO_INPUT);         //  Rotary Encoder B
    gpio_configure(20, GPIO_OUTPUT);        //  PN TX
    gpio_configure(21, GPIO_OUTPUT);        //  ATT
    gpio_configure(22, GPIO_INPUT);         //  dat5
    gpio_configure(23, GPIO_INPUT);         //  dat4
    gpio_configure(24, GPIO_INPUT);         //  dat3
    gpio_configure(25, GPIO_INPUT);         //  dat0
    gpio_configure(26, GPIO_OUTPUT);        //  PN RX
    gpio_configure(27, GPIO_INPUT);         //  dat6
    
    gpio_configure_pull(4, GPIO_PULLDOWN);
    gpio_configure_pull(5, GPIO_PULLDOWN);
    gpio_configure_pull(9, GPIO_PULLDOWN);
    gpio_configure_pull(10, GPIO_PULLDOWN);
    gpio_configure_pull(13, GPIO_PULLDOWN);
    gpio_configure_pull(14, GPIO_PULLDOWN);
    gpio_configure_pull(15, GPIO_PULLDOWN);
    gpio_configure_pull(17, GPIO_PULLDOWN);
    gpio_configure_pull(18, GPIO_PULLDOWN);
    gpio_configure_pull(19, GPIO_PULLDOWN);
    gpio_configure_pull(22, GPIO_PULLDOWN);
    gpio_configure_pull(23, GPIO_PULLDOWN);
    gpio_configure_pull(24, GPIO_PULLDOWN);
    gpio_configure_pull(25, GPIO_PULLDOWN);
    gpio_configure_pull(27, GPIO_PULLDOWN);
    

}

void FFT(double* real, double* imag, int length, int sw)
{
    //　IQによる帯域2倍に対応
    
    int i, j, k, l, half;
    double m, h, g, q, p, a, b, phase;
    double swap;
    
    half=length/2;
    
    //*********************** 波形読込と同時にハミング窓を掛ける **********************
    
    for (i=0; i<=length; i++) {
        
        phase=2*M_PI*(float)i/(float)length;                // 0 to 2*pi
        
        if (sw==0) {
            a=1.414*real[i]*(0.50-0.50*cosf(phase));        //Hanning
            b=1.414*imag[i]*(0.50-0.50*cosf(phase));        //Hanning
        }
        if (sw==1) {
            a=1.669*real[i]*(0.35875-0.48829*cosf(phase)+0.14128*cosf(2*phase)-0.01168*cosf(3*phase));  //Blackman-Harris
            b=1.669*imag[i]*(0.35875-0.48829*cosf(phase)+0.14128*cosf(2*phase)-0.01168*cosf(3*phase));  //Blackman-Harris
        }
        if (sw==2) {
            a=1.000*real[i];
            b=1.000*imag[i];
        }
        
        real[i]=a;
        imag[i]=b;
    }
    
    //******************************* FFTメイン **********************************
    m = log2(length);
    
    l = length;
    h = 1;
    
    for (g=1; g<=m; g++) {
        l = l / 2;
        k = 0;
        for (q=1; q<=h; q++) {
            p = 0;
            for (i=k; i<=l+k-1; i++) {
                j = i + l;
                a = real[i] - real[j];
                b = imag[i] - imag[j];
                real[i] = real[i] + real[j];
                imag[i] = imag[i] + imag[j];
                if (p==0){
                    real[j] = a;
                    imag[j] = b;
                }else{
                    real[j] = a * cosf(2 * M_PI * p / (float)length) + b * sinf(2 * M_PI * p / (float)length);
                    imag[j] = b * cosf(2 * M_PI * p / (float)length) - a * sinf(2 * M_PI * p / (float)length);
                }
                p = p + h;
            }
            k = k + l + l;
        }
        h = h + h;
    }
    
    j = length / 2;
    
    for (i=1; i<=length-1; i++) {
        k = length;
        if (j<i){
            swap = real[i];
            real[i] = real[j];
            real[j] = swap;
            swap = imag[i];
            imag[i] = imag[j];
            imag[j] = swap;
        }
        
        do{
            k = k / 2;
            j = j - k;
            if (k==0 && j==0) break;
        }while ((j+k)>=k);
        
        j = j + k + k;
    }
    
}


void iirlpf2coef(int n, int fc, int smplrate)
{
    //-------------------- IIR フィルタ係数生成 ------------------------
    int ws = smplrate;      //サンプリング周波数
    int w1;                 //カットオフ周波数
    float w;
    
    w1=fc;
    w = tanf(M_PI*w1/ws);
    
    a1[n] = -(2 * w*w - 2) / (1 + sqrtf(2) * w + w*w);
    a2[n] = -(1 - sqrtf(2) * w + w*w) / (1 + sqrtf(2) * w + w*w);
    b0[n] = w*w / (1 + sqrtf(2) * w + w*w);
    b1[n] = (2 * w*w) / (1 + sqrtf(2) * w + w*w);
    b2[n] = w*w / (1 + sqrtf(2) * w + w*w);
    //---------------------------------------------------------------
    
}

void iirhpf2coef(int n, int fc, int smplrate)
{
    //-------------------- IIR フィルタ係数生成 ------------------------
    int ws = smplrate;      //サンプリング周波数
    int w1;                 //カットオフ周波数
    float w;
    
    w1=fc;
    w = tanf(M_PI*w1/ws);
    
    a1[n] = -(2 * w - 2) / (1 + sqrtf(2)*w + w*w);
    a2[n] = -(1 - sqrtf(2)*w + w*w) / (1 + sqrtf(2)*w + w*w);
    b0[n] = 1 / (1 + sqrtf(2)*w + w*w);
    b1[n] = -2 / (1 + sqrtf(2)*w + w*w);
    b2[n] = 1 / (1 + sqrtf(2)*w + w*w);
    //---------------------------------------------------------------
    
}

void iirlpf1coef(int n, int fc, int smplrate)
{
    //-------------------- IIR フィルタ係数生成 ------------------------
    int ws = smplrate;      //サンプリング周波数
    int w1;                 //カットオフ周波数
    float w;
    
    w1=fc;
    w = tanf(M_PI*w1/ws);
    
    a1[n] = (1 - w) / (1 + w);
    a2[n] = 0;
    b0[n] = 1 / (1 + w);
    b1[n] = 1 / (1 + w);
    b2[n] = 0;
    //---------------------------------------------------------------
    
}

//-------------------------------------------------------------------
//                          IIR Filter
//-------------------------------------------------------------------
float iir(int n, float vin)
{
    static float z21[6];
    static float z22[6];
    float vo;
    
    vin = vin + a1[n] * z21[n];
    vin = vin + a2[n] * z22[n];
    vo = b0[n] * vin;
    vo = vo + b1[n] * z21[n];
    vo = vo + b2[n] * z22[n];
    z22[n] = z21[n];
    z21[n] = vin;
    
    return vo;
}

//-------------------------------------------------------------------
//                            FIR係数の生成
//-------------------------------------------------------------------
void Fircoef(double* fcoef, int fc, int length)
{
    double fr[256];
    double fj[256];
    double gros;
    int n;
    int half;
    
    half = length/2;
    
    //  周波数特性を定義
    for (n=0; n<half; n++) {
        if (n<fc) {
            fr[n]=1;
            fr[length-n-1]=1;
            fj[n]=1;
            fj[length-n-1]=1;
        }else{
            fr[n]=0;
            fr[length-n-1]=0;
            fj[n]=0;
            fj[length-n-1]=0;
        }
    }
    
    FFT(fr, fj, length, 2);         //　窓関数は掛けない(2)
    
    gros=0;
    
    //  係数をシフトする
    for (n=0; n<half; n++) {
        fcoef[half+n]   =fr[n];
        gros=gros+fr[n];
    }
    for (n=half; n<length; n++) {
        fcoef[n-half]   =fr[n];
        gros=gros+fr[n];
    }
    
    //　正規化
    for (n=0; n<length; n++) {
        fcoef[n]=fcoef[n]/gros;
    }
    
    //    for (n=0; n<length; n++) {
    //        printf( "FFT coef %d %7.5f \n",n,fcoef[n] );
    //    }
    
}

//-------------------------------------------------------------------
//                            Hilbert係数の生成
//-------------------------------------------------------------------
void Hilcoef(double* fcoef, int fc, int length)
{
    double fr[256];
    double fj[256];
    double gros;
    int n;
    int half;
    
    half = length/2;
    
    //  周波数特性を定義
    for (n=0; n<half; n++) {
        if (n<fc) {
            fr[n]=1;
            fr[length-n-1]=1;
            fj[n]=1;
            fj[length-n-1]=-1;
        }else{
            fr[n]=0;
            fr[length-n-1]=0;
            fj[n]=0;
            fj[length-n-1]=0;
        }
    }
    
    FFT(fr, fj, length, 2);         //　窓関数は掛けない(2)
    
    gros=0;
    
    //  係数をシフトする
    for (n=0; n<half; n++) {
        fcoef[half+n]   =fr[n];
        gros=gros+fr[n];
    }
    for (n=half; n<length; n++) {
        fcoef[n-half]   =fr[n];
        gros=gros+fr[n];
    }
    
    //　正規化
    for (n=0; n<length; n++) {
        fcoef[n]=fcoef[n]/gros;
    }
    
    //    for (n=0; n<length; n++) {
    //        printf( "FFT coef %d %7.5f \n",n,fcoef[n] );
    //    }
    
}

unsigned int * get_base_address()
{
    int fd = open("/dev/mem", O_RDWR | O_SYNC);
    if(fd < 0){
        printf("cannot open /dev/mem\n"); exit(-1);
    }
    
    #define PAGE_SIZE (4096)
    void * mmaped = mmap(NULL,
                         PAGE_SIZE,
                         PROT_READ | PROT_WRITE,
                         MAP_SHARED,
                         fd,
                         0x3F200000);
    if (mmaped < 0) { printf("mmap failure\n");
        exit(-1); }
    
    close(fd);
    return (unsigned int *) mmaped;
    
}

void gpio_init()
{
    if (Gpio) return;
    int fd;
    void *gpio_map;
    fd = open("/dev/mem", O_RDWR | O_SYNC);
    gpio_map = mmap(NULL, BLOCK_SIZE,
                    PROT_READ | PROT_WRITE, MAP_SHARED,
                    fd, GPIO_BASE );
    close(fd);
    Gpio = (unsigned int *) gpio_map;
}

void gpio_configure(int pin, int mode)
{
    int index = pin / 10;
    unsigned int mask = ~(0x7 << ((pin % 10) * 3));
    Gpio[index] = (Gpio[index] & mask) | ((mode & 0x7) << ((pin % 10) * 3));
}

void gpio_configure_pull (int pin, int pullmode)
{
    Gpio[37] = pullmode & 0x3;
    usleep(1);
    Gpio[38] = 0x1 << pin;
    usleep(1);
    Gpio[37] = 0;
    Gpio[38] = 0;
}

void gpio_set(int pin)
{
    Gpio[7] = 0x1 << pin;
}

void gpio_clear(int pin)
{
    Gpio[10] = 0x1 << pin;
}

int gpio_read (int pin)
{
    return (Gpio[13] & (0x1 << pin)) != 0;
}

// ウィンドウにタイトルを設定する 
void SetTitle(Display *display, Window window, char *title)
{
	XTextProperty ct;
	
	XmbTextListToTextProperty(display, &title, 1, XCompoundTextStyle, &ct);
	XSetWMName(display, window, &ct);
}

//　SPAN変更に伴うGPIO設定
void nw(int sw)
{
    if (sw==0){     // Narrow 10kHz
        gpio_clear(8);
        gpio_clear(16);
        zoom=0.2049;
        zosc=0.4766;
    }
    if (sw==1){     // Narrow 25kHz
        gpio_clear(8);
        gpio_clear(16);
        zoom=0.5123;
        zosc=0.2383;
    }
    if (sw==2){     // Narrow 50kHz
        gpio_clear(8);
        gpio_clear(16);
        zoom=1.0250;
        zosc=0.0953;
    }
    if (sw==3){     // Mid 0.2MHz
        gpio_set(8);
        gpio_set(16);
        zoom=0.1282;
        zosc=0.6094;
    }
    if (sw==4){     // Mid 0.5MHz
        gpio_set(8);
        gpio_set(16);
        zoom=0.3205;
        zosc=0.3047;
    }
    if (sw==5){     // Mid 1.0MHz
        gpio_set(8);
        gpio_set(16);
        zoom=0.6410;
        zosc=0.1524;
    }
    if (sw==6){     // Wide 5MHz
        gpio_set(8);
        gpio_clear(16);
        zoom=0.1000;
        zosc=0.9766;
    }
    if (sw==7){     // Wide 10MHz
        gpio_set(8);
        gpio_clear(16);
        zoom=0.2000;
        zosc=0.4883;
    }
    if (sw==8){     // Wide 25MHz
        gpio_set(8);
        gpio_clear(16);
        zoom=0.5000;
        zosc=0.1953;
    }
}

//　ATT設定
void attset(int sw)
{
    if (sw==0) {      // ATT=0dB
        gpio_clear(21);
    }
    if (sw==1) {      // ATT=12dB
        gpio_set(21);
    }
}

//　トップスイッチサブウィンドウの表示
void TopSW()
{
    char txt[20];
    int x;
    int fx, fy;
    int top_w=480;
    
    //　画面のクリア
    XClearWindow(dpy, wint);
    
    // START
    x=0;
    sprintf(txt, "START");
    if (sw_start==1) XSetForeground( dpy, gc, red   );
    if (sw_start==0) XSetForeground( dpy, gc, white );
    XDrawString(dpy, wint, gc, 10+top_w*x/10, 12, txt, 5);
    
    // STOP
    x=1;
    sprintf(txt, "STOP");
    if (sw_start==1) XSetForeground( dpy, gc, white );
    if (sw_start==0) XSetForeground( dpy, gc, red   );
    XDrawString(dpy, wint, gc, 13+top_w*x/10, 12, txt, 4);
    
    // DEMOD
    x=2;
    sprintf(txt, "DEMOD");
    if (dem==0) XSetForeground( dpy, gc, white );
    if (dem==1) XSetForeground( dpy, gc, red   );
    XDrawString(dpy, wint, gc, 10+top_w*x/10, 12, txt, 5);
    
    // MEMIN
    x=3;
    sprintf(txt, "MEMIN");
    if (sw_memin==0) XSetForeground( dpy, gc, white );
    if (sw_memin==1) XSetForeground( dpy, gc, red   );
    XDrawString(dpy, wint, gc, 10+top_w*x/10, 12, txt, 5);
    
    // MEM1
    x=4;
    sprintf(txt, "MEM1");
    XSetForeground( dpy, gc, aqua );
    XDrawString(dpy, wint, gc, 13+top_w*x/10, 12, txt, 4);
    
    // MEM2
    x=5;
    sprintf(txt, "MEM2");
    XSetForeground( dpy, gc, aqua );
    XDrawString(dpy, wint, gc, 13+top_w*x/10, 12, txt, 4);
    
    // MEM3
    x=6;
    sprintf(txt, "MEM3");
    XSetForeground( dpy, gc, aqua );
    XDrawString(dpy, wint, gc, 13+top_w*x/10, 12, txt, 4);
    
    // COPY
    x=7;
    sprintf(txt, "COPY");
    XSetForeground( dpy, gc, white );
    XDrawString(dpy, wint, gc, 13+top_w*x/10, 12, txt, 4);
    
    // REC
    x=8;
    sprintf(txt, "REC");
    if (rcd==0) XSetForeground( dpy, gc, white );
    if (rcd==1) XSetForeground( dpy, gc, red );
    XDrawString(dpy, wint, gc, 16+top_w*x/10, 12, txt, 3);
    
    // OFF
    x=9;
    sprintf(txt, "OFF");
    XSetForeground( dpy, gc, white );
    XDrawString(dpy, wint, gc, 16+top_w*x/10, 12, txt, 3);
    
}

//　パラメータ設定サブウィンドウの表示
void DispControl()
{
    char txt[55];
    int x;
    int fx, fy;
    int stpy=17;        // Text Step Y
    
    //　画面のクリア
    XClearWindow(dpy, winp);
    
    // Function
    sprintf(txt, "->");
    XSetForeground( dpy, gc, red );
    if (func>= 0 && func<10) fx=5;
    if (func>=10 && func<20) fx=185;
    if (func>=20 && func<30) fx=335;
    fy=20+(func-int(func/10)*10)*stpy;
    XDrawString(dpy, winp, gc, fx, fy, txt, 2);
        
    // FREQ
    //if (vf>0) fpll=vf;
    sprintf(txt, "Center=%9.4f MHz", fpll);
    XSetForeground( dpy, gc, white ); 
    XDrawString(dpy, winp, gc, 20, 20, txt, 20);

	// DIGIT Mark
	if (fstep>0.5) x=80-6*log10(fstep);
	if (fstep<0.5) x=86-6*log10(fstep);
    sprintf(txt, "_");
    XSetForeground( dpy, gc, red ); 
    XDrawString(dpy, winp, gc, x, 23, txt, 1);

    // SPAN/Time
    if (spa==1 && nwsw==0) sprintf(txt, "Span= 10.0 kHz   ");
    if (spa==1 && nwsw==1) sprintf(txt, "Span= 25.0 kHz   ");
    if (spa==1 && nwsw==2) sprintf(txt, "Span= 50.0 kHz   ");
    if (spa==1 && nwsw==3) sprintf(txt, "Span=  0.2 MHz   ");
    if (spa==1 && nwsw==4) sprintf(txt, "Span=  0.5 MHz   ");
    if (spa==1 && nwsw==5) sprintf(txt, "Span=  1.0 MHz   ");
    if (spa==1 && nwsw==6) sprintf(txt, "Span=  5.0 MHz   ");
    if (spa==1 && nwsw==7) sprintf(txt, "Span= 10.0 MHz   ");
    if (spa==1 && nwsw==8) sprintf(txt, "Span= 25.0 MHz   ");
    
    if (spa==0 && nwsw==0) sprintf(txt, "Span=  5.0 ms/div");
    if (spa==0 && nwsw==1) sprintf(txt, "Span=  2.0 ms/div");
    if (spa==0 && nwsw==2) sprintf(txt, "Span=  1.0 ms/div");
    if (spa==0 && nwsw==3) sprintf(txt, "Span=  0.2 ms/div");
    if (spa==0 && nwsw==4) sprintf(txt, "Span=  0.1 ms/div");
    if (spa==0 && nwsw==5) sprintf(txt, "Span= 50.0 us/div");
    if (spa==0 && nwsw==6) sprintf(txt, "Span= 10.0 us/div");
    if (spa==0 && nwsw==7) sprintf(txt, "Span=  5.0 us/div");
    if (spa==0 && nwsw==8) sprintf(txt, "Span=  2.0 us/div");
    
    XSetForeground( dpy, gc, white );
    XDrawString(dpy, winp, gc, 200, 20, txt, 17);
    
    // Function
    if (spa==0) sprintf(txt, "FUNC    Oscillo ");
    if (spa==1) sprintf(txt, "FUNC    Spectrum");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20, txt, 16);
    
    // Scale dB/div
    XSetForeground( dpy, gc, white );
    sprintf(txt, "Scale =%8.3f  dB/div", (float)sd);
    XDrawString(dpy, winp, gc, 20, 20+stpy*1, txt, 23);
    
    // Ref Level
    XSetForeground( dpy, gc, white );
    sprintf(txt, "Ref =%5.1f dBm", (float)sr);
    XDrawString(dpy, winp, gc, 200, 20+stpy*1, txt, 14);
    
    // Oscillo CH1
    XSetForeground( dpy, gc, aqua );
    sprintf(txt, "CH1 (I)   %5.3f  mV/div", vd1);
    XDrawString(dpy, winp, gc, 20, 20+stpy*2, txt, 23);
    sprintf(txt, "POS = %4.1f div", vo1);
    XDrawString(dpy, winp, gc, 200, 20+stpy*2, txt, 14);
    
    // Oscillo CH2
    XSetForeground( dpy, gc, fuchsia );
    sprintf(txt, "CH2 (Q)   %5.3f  mV/div", vd2);
    XDrawString(dpy, winp, gc, 20, 20+stpy*3, txt, 23);
    sprintf(txt, "POS = %4.1f div", vo2);
    XDrawString(dpy, winp, gc, 200, 20+stpy*3, txt, 14);
    
    // Oscillo CH3
    XSetForeground( dpy, gc, green );
    sprintf(txt, "CH3       %5.3f  mV/div", vd3);
    XDrawString(dpy, winp, gc, 20, 20+stpy*4, txt, 23);
    sprintf(txt, "POS = %4.1f div", vo3);
    XDrawString(dpy, winp, gc, 200, 20+stpy*4, txt, 14);
    
    // FIR
    XSetForeground( dpy, gc, white );
    sprintf(txt, "FIR BW=%6.1f    kHz", (float)bwf);
    XDrawString(dpy, winp, gc, 20, 20+stpy*5, txt, 20);
    
    // IIR
    XSetForeground( dpy, gc, white );
    sprintf(txt, "IIR BW=%6.1f    kHz", (float)bws);
    XDrawString(dpy, winp, gc, 20, 20+stpy*6, txt, 20);
    
    // GAIN
    XSetForeground( dpy, gc, white );
    sprintf(txt, "GAIN  =%6.1f    dB ", (float)ifg);
    XDrawString(dpy, winp, gc, 20, 20+stpy*7, txt, 20);
    
    // SQL
    XSetForeground( dpy, gc, white );
    sprintf(txt, "SQL =%5.0f dBm", sqv);
    XDrawString(dpy, winp, gc, 200, 20+stpy*5, txt, 14);
    
    // Average
    XSetForeground( dpy, gc, white );
    sprintf(txt, "AVE =%5d", aven);
    XDrawString(dpy, winp, gc, 200, 20+stpy*6, txt, 10);
    
    //  DcCAL
    if (ofc==0) sprintf(txt, "DcCAL   OFF");
    if (ofc==1) sprintf(txt, "DcCAL   ON ");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20+stpy*1, txt, 11);
    
    //  Emphasis
    if (emp==0) sprintf(txt, "Empha   OFF");
    if (emp==1) sprintf(txt, "Empha   ON ");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20+stpy*2, txt, 11);
    
    //  XY screen
    if (xys==0) sprintf(txt, "XY      OFF");
    if (xys==1) sprintf(txt, "XY      ON ");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20+stpy*3, txt, 11);
    
    //  MODE
    if (mode==0) sprintf(txt, "MODE    FM ");
    if (mode==1) sprintf(txt, "MODE    AM ");
    if (mode==2) sprintf(txt, "MODE    DSB");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20+stpy*4, txt, 11);
    
    //  CH3 Input
    if (afs==0) sprintf(txt, "CH3 in  OFF");
    if (afs==1) sprintf(txt, "CH3 in  AF ");
    if (afs==2) sprintf(txt, "CH3 in  POW");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20+stpy*5, txt, 11);
    
    //  Scan
    if (scansw==0) sprintf(txt, "SCAN    OFF ");
    if (scansw==1) sprintf(txt, "SCAN    HOP ");
    if (scansw==2) sprintf(txt, "SCAN    SNGL");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20+stpy*6, txt, 12);
    
    //  ATT
    if (att==0) sprintf(txt, "ATT     0dB ");
    if (att==1) sprintf(txt, "ATT     12dB");
    XSetForeground( dpy, gc, gray );
    XDrawString(dpy, winp, gc, 350, 20+stpy*7, txt, 12);
    
}

//　任意の色を合成
unsigned long RGB(Display *display,
                  unsigned char r,
                  unsigned char g,
                  unsigned char b)
{
    Colormap cm;
    XColor xc;
    
    cm = DefaultColormap(display, DefaultScreen(display));
    
    xc.red = 257 * r;
    xc.green = 257 * g;
    xc.blue = 257 * b;
    XAllocColor(display, cm, &xc);
    
    return xc.pixel;
}

//　f特補正データの読み込み
void getFresp(double* fresp)
{
    FILE *fp;                   //  fresp file pointer
    double fr[100][2];
    char sl[64];
    char *pm[2] ;
    int n, m, f;
    int f1, f2;
    float g0, g1, g2, re;
    
    if ((fp=fopen("fresp.txt", "r")) == NULL) {
        printf("Not exist fresp.txt \n");
        for (f=50; f<=2200; f++) {
            fresp[f]=1;         // Dummy data when no file
        }
        return;
    }else{
        printf("Load fresp.txt OK!\n");
    }
    
    n=0;
    while (fgets(sl, sizeof(sl), fp) != NULL) {
        pm[0] = strtok(sl,",");
        pm[1] = strtok(NULL,",");
        fr[n][0]=atof(pm[0]);
        fr[n][1]=atof(pm[1]);
        n++;
    }
    
    //　f特補完
    for (m=0; m<n-1; m++) {
        f1=int(fr[m  ][0]);     // freq lower
        f2=int(fr[m+1][0]);     // freq upper
        g1=    fr[m  ][1];      // gain lower
        g2=    fr[m+1][1];      // gain upper
        
        for (f=f1; f<f2; f++) {
            g0=g1+(float)(f-f1)*(g2-g1)/(float)(f2-f1);     // Fitting
            re=pow(10,g0/20);                               // Conv dB to Linear
            fresp[f]=re;                                    // Move to Table
        }
    }
    
    fclose(fp);
}

//　CICフィルタ補正f特カーブ生成
void cicFresp(double* cicw, double* cicn)
{
    int f;
    int fs;
    int fw;
    double g1, g2;
    
    //　FFT出力に合わせるため±fを入れ替えてあるので注意
    
    // CIC Wide = 1024
    fs=1024;
    fw=512;
    for (f=1; f<=fw; f++) {
        g1=(sinf(M_PI*(float)f/(float)fs))/(M_PI*(float)f/(float)fs);
        g2=g1*g1*g1*g1;
        cicw[fs-f]=g2;
        cicw[   f]=g2;
    }
    cicw[0]=1;
    
    // CIC Wide = 512
    fs=512;
    fw=256;
    for (f=1; f<=fw; f++) {
        g1=(sinf(M_PI*(float)f/(float)fs))/(M_PI*(float)f/(float)fs);
        g2=g1*g1*g1*g1;
        cicn[fs-f]=g2;
        cicn[   f]=g2;
    }
    cicn[0]=1;
}

// ピクスマップにグリッドライン描画
void pixwrite(int W, int H, int sd, int sr, int spa, float SQL)
{
    int n;
    int mkrsq;
    char txt[20];

    // ピクスマップ背景生成＆画面クリア
    XSetForeground(dpy, g_gc, RGB(dpy, 10, 50, 50));
    XFillRectangle(dpy, pixgrid, g_gc, 0, 0, W, H);
    
    
    //　グリッドライン描画
    XSetForeground( dpy, g_gc, RGB(dpy, 10 ,20 ,20));
    for (n=0; n<11; n++){
        XDrawLine( dpy, pixgrid, g_gc, 0, n*H/10, W, n*H/10);
        XDrawLine( dpy, pixgrid, g_gc, n*W/10, 0, n*W/10, H);
    }
    
    if (spa==0) return;
    
    //  チューニング周波数（センターライン）
    XSetForeground( dpy, g_gc, RGB(dpy, 100 ,0 ,0));
    XDrawLine( dpy, pixgrid, g_gc, W/2, 0, W/2, H);
    
    //  スケルチレベルマーカー
    mkrsq=H*(sr-SQL)/(10*sd);
    XSetForeground( dpy, g_gc, RGB(dpy, 200 ,0 ,0));
    XDrawLine( dpy, pixgrid, g_gc, 0, mkrsq, W, mkrsq);
    
    //  SQL Marker Line
    sprintf(txt, "  SQL=%5.0f dBm",SQL);
    XSetForeground( dpy, g_gc, RGB(dpy, 200 ,0 ,0));
    XDrawString(dpy, pixgrid, g_gc, 5, mkrsq-5, txt, 15);

}

//　設定内容をファイルに保存
void savemem(int ch)
{
    FILE *fp;                       //  data out file pointer
    char sdata[128];
    
    if (ch==1) fp=fopen("mem1.txt", "w");
    if (ch==2) fp=fopen("mem2.txt", "w");
    if (ch==3) fp=fopen("mem3.txt", "w");
    
    sprintf(sdata, "Curr Freq  , %8.3f \n", fpll);
    fputs(sdata, fp);
    sprintf(sdata, "SPA dB/div , %5d   \n", sd);
    fputs(sdata, fp);
    sprintf(sdata, "SPA Ref dBm, %5d   \n", sr);
    fputs(sdata, fp);
    sprintf(sdata, "FIR BW kHz , %5d   \n", bwf);
    fputs(sdata, fp);
    sprintf(sdata, "IIR BW kHz , %5d   \n", bws);
    fputs(sdata, fp);
    sprintf(sdata, "Average    , %5d   \n", aven);
    fputs(sdata, fp);
    sprintf(sdata, "SQL dBm    , %8.3f \n", sqv);
    fputs(sdata, fp);
    sprintf(sdata, "CH1 mV/div , %8.3f \n", vd1);
    fputs(sdata, fp);
    sprintf(sdata, "CH2 mV/div , %8.3f \n", vd2);
    fputs(sdata, fp);
    sprintf(sdata, "CH3 mV/div , %8.3f \n", vd3);
    fputs(sdata, fp);
    sprintf(sdata, "CH1 pos/div, %8.3f \n", vo1);
    fputs(sdata, fp);
    sprintf(sdata, "CH2 pos/div, %8.3f \n", vo2);
    fputs(sdata, fp);
    sprintf(sdata, "CH3 pos/div, %8.3f \n", vo3);
    fputs(sdata, fp);
    sprintf(sdata, "Span N/M/W , %1d   \n", nwsw);
    fputs(sdata, fp);
    sprintf(sdata, "Mode F/A/D , %1d   \n", mode);
    fputs(sdata, fp);
    sprintf(sdata, "DC off CAL , %1d   \n", ofc);
    fputs(sdata, fp);
    sprintf(sdata, "Emphasis   , %1d   \n", emp);
    fputs(sdata, fp);
    sprintf(sdata, "Scan SW    , %1d   \n", scansw);
    fputs(sdata, fp);
    sprintf(sdata, "Spectrum   , %1d   \n", spa);
    fputs(sdata, fp);
    sprintf(sdata, "Demod mode , %1d   \n", dem);
    fputs(sdata, fp);
    sprintf(sdata, "CH3 input  , %1d   \n", afs);
    fputs(sdata, fp);
    sprintf(sdata, "XY Oscillo , %1d   \n", xys);
    fputs(sdata, fp);
    sprintf(sdata, "IF Gain    , %1d   \n", ifg);
    fputs(sdata, fp);
    
    sw_memin=0;
    
    fclose(fp);
}

//　設定内容をファイルから読み込み
void loadmem(int ch)
{
    FILE *fp;                       //  data out file pointer
    char sdata[128];
    char *item;
    char *param;
    int n;
    
    if (ch==1) fp=fopen("mem1.txt", "r");
    if (ch==2) fp=fopen("mem2.txt", "r");
    if (ch==3) fp=fopen("mem3.txt", "r");
    
    if (fp== NULL) {
        printf("Not exist Memory file!\n");
        return;
    }
    
    n=0;
    while (fgets(sdata, sizeof(sdata), fp) != NULL) {
        item  = strtok(sdata,",");
        param = strtok(NULL,",");
        if (n== 0) fpll     =atof(param);
        if (n== 1) sd       =atoi(param);
        if (n== 2) sr       =atoi(param);
        if (n== 3) bwf      =atoi(param);
        if (n== 4) bws      =atoi(param);
        if (n== 5) aven     =atoi(param);
        if (n== 6) sqv      =atof(param);
        if (n== 7) vd1      =atof(param);
        if (n== 8) vd2      =atof(param);
        if (n== 9) vd3      =atof(param);
        if (n==10) vo1      =atof(param);
        if (n==11) vo2      =atof(param);
        if (n==12) vo3      =atof(param);
        if (n==13) nwsw     =atoi(param);
        if (n==14) mode     =atoi(param);
        if (n==15) ofc      =atoi(param);
        if (n==16) emp      =atoi(param);
        if (n==17) scansw   =atoi(param);
        if (n==18) spa      =atoi(param);
        if (n==19) dem      =atoi(param);
        if (n==20) afs      =atoi(param);
        if (n==21) xys      =atoi(param);
        if (n==22) ifg      =atoi(param);
        n++;
    }
    
    dem=0;                  //　必ず復調モードOFFから
    spa=1;                  //　必ずスペアナモードから
    nwsw=5;                 //　必ずBW=1.56Mから
    
    fhome=fpll;

    fclose(fp);
}

//　画面１枚分の波形をCSVファイルに保存
void copy(double* av, double* ch1, double* ch2, double* ch3, int scope_h)
{
    FILE *fp;                       //  data out file pointer
    char sdata[128];
    int n, length;
    float f, bw, db, t, ts, am;
    float c1, c2, c3;
    
    if (nwsw<=2) {
        bw=0.048828125;
        ts=0.01048576;
        am=0.1;
    }
    if (nwsw>2 && nwsw<=5) {
        bw=1.562500000;
        ts=0.00032768;
        am=0.1;
    }
    if (nwsw>5) {
        bw=50.00000000;
        ts=0.00001024;
        am=0.1;
    }
    
    fp=fopen("copy.csv", "w");
    
    //　オシロ画面のコピー
    if (spa==0) {
        length=512;
        sprintf(sdata, "T(s),CH1(mV),CH2(mV),CH3(mV)\n");
        fputs(sdata, fp);
        for (n=1; n<length; n++) {
            t=ts*((float)n-1.0)/512.0;
            c1=-(ch1[n]/(float)scope_h+vo1/10.0-0.5)*vd1/am;
            c2=-(ch2[n]/(float)scope_h+vo2/10.0-0.5)*vd2/am;
            c3=-(ch3[n]/(float)scope_h+vo3/10.0-0.5)*vd3/am;
            sprintf(sdata, "%10.9f,%8.5f,%8.5f,%8.5f\n", t, c1, c2, c3);
            fputs(sdata, fp);
        }
    }
    
    //　スペアナ画面のコピー　BW=48.8kHz
    if (spa==1 && nwsw<=2) {
        length=512;
        sprintf(sdata, "Freq(MHz),Level(dBm)\n");
        fputs(sdata, fp);
        for (n=257; n<length; n++) {
            f=fpll-(bw/2)+(n-256)*(bw/2)/256;
            db=(-av[n]+sr*((float)scope_h/10.0)/sd)/(((float)scope_h/10.0)/sd);
            sprintf(sdata, "%12.6f,%8.5f\n", f, db);
            fputs(sdata, fp);
        }
        for (n=1; n<=256; n++) {
            f=fpll+(n-1)*(bw/2)/256;
            db=(-av[n]+sr*((float)scope_h/10.0)/sd)/(((float)scope_h/10.0)/sd);
            sprintf(sdata, "%12.6f,%8.5f\n", f, db);
            fputs(sdata, fp);
        }
    }
    
    //　スペアナ画面のコピー　BW=1.56MHz & 50MHz
    if (spa==1 && nwsw>2) {
        length=1024;
        sprintf(sdata, "Freq(MHz),Level(dBm)\n");
        fputs(sdata, fp);
        for (n=513; n<length; n++) {
            f=fpll-(bw/2)+(n-512)*(bw/2)/512;
            db=(-av[n]+sr*((float)scope_h/10.0)/sd)/(((float)scope_h/10.0)/sd);
            sprintf(sdata, "%12.6f,%8.5f\n", f, db);
            fputs(sdata, fp);
        }
        for (n=1; n<=512; n++) {
            f=fpll+(n-1)*(bw/2)/512;
            db=(-av[n]+sr*((float)scope_h/10.0)/sd)/(((float)scope_h/10.0)/sd);
            sprintf(sdata, "%12.6f,%8.5f\n", f, db);
            fputs(sdata, fp);
        }
    }
    
    fclose(fp);
}

